/* Converts a pair of alleles into an integer code   */
/* a==1st allele, b==2nd allele, c==tot # of alleles */

int code_dna (int a, int b, int c)
{
   int   temp;

   if (a > b)
   {
      temp = a;
      a = b;
      b = temp;
   }
   if ((a + b) == 0)
      return 0;
   if (a == 1)
      return b;
   else
      return ((a - 1) * (2 * c - a + 2) * .5 + b - a + 1);
}

/* Converts an integer coded genotype into a pair   */
/* of integer coded alleles.                        */

void uncode_dna (int g, int *a, int *b, int c)
{
   int   left, right, count, flag;

   if (g)
   {
      left = 1;
      right = c;
      count = 0;
      flag = 0;
      while (count <= c && flag == 0)
      {
         count++;
         if (left <= g && g <= right)
         {
            flag = 1;
            *a = count;                 /* the 'row' allele */
            if (count == 1)
               *b = g;                  /* the 'col' allele */
            else
               *b = g + count - 1 - ((count - 1) * (2 * c - count + 2) * .5);
         }
         left = right + 1;
         right = left + c - count - 1;
      }
   }
   else
   {
      *a = 0;
      *b = 0;
   }
}

/* Returns the segregation probability from parent
   genos: (par1, par2) to offspring geno: (off), and 
   k is the total number of alleles                   */

double sege (int par1, int par2, int off, int k)
{
   int   a1, a2, a3, a4, a5, a6;
   double z1, z2, seg;

   uncode_dna (par1, &a1, &a2, k);
   uncode_dna (par2, &a3, &a4, k);
   uncode_dna (off, &a5, &a6, k);

   /* phase the offspring (a5|a6) */
   z1 = .5;
   z2 = .5;
   if (a5 != a1 && a5 != a2)
      z1 = 0;
   if (a6 != a3 && a6 != a4)
      z2 = 0;
   if (a5 == a1 && a5 == a2)
      z1 = 1;
   if (a6 == a3 && a6 == a4)
      z2 = 1;
   seg = z1 * z2;

   /* phase is unambiguous  */
   if (a5 == a6)
      return (seg);

   else
   {
      /* phase the offspring (a6|a5) */
      z1 = .5;
      z2 = .5;
      if (a6 != a1 && a6 != a2)
         z1 = 0;
      if (a5 != a3 && a5 != a4)
         z2 = 0;
      if (a6 == a1 && a6 == a2)
         z1 = 1;
      if (a5 == a3 && a5 == a4)
         z2 = 1;
      seg = seg + (z1 * z2);
   }
   return (seg);
}

/* This function could be used to fill pene_mat[i][][] */
/* with different error models. However, since the     */
/* present implementation uses the simplest of error   */
/* models, this function isn't used yet (08/31/07).    */

double pene (int g, int y, double eps, int k)
{
   k = k * (k + 1) / 2;
   if (y == 0)
      return 1;
   if (g == y)
      return (1 - eps);
   return (eps / (k - 1));
}

/* All locus_peel() calculations are locus specific.                          */
/* The peeling aspects of locus_peel() are structured around ea. individual's */
/* position in the vector invol[]. Whereas, the reverse peeling aspects of    */
/* locus_peel() are structured around each individual's id.                   */
/* While the variable names im and ife suggest that mothers should come       */
/* before fathers in the peeling sequence, there input order is irrelevant.   */
/* locus_peel() returns a double, which is the P(marker phenos at locus j).   */
/* Depending on the value of rvsp_flag, the genotypes at that locus may or    */
/* may not be sampled conditional on the marker phenos.                       */

double locus_peel (int rp_flg, double eps, int fam_numb, int *seq,
                   int wloc, int ***pgif, int *sf, double **pene_mat,
                   double ***sege_mat)
{
   int   off, cp_fou[NP], upp[NP][3], cluster[4][NST], upp_count = 0;
   int   jj, jjj, pivot, itru_geno[NP], par1, par2, pp, cend, rlen, ii;
   int   kk, kkk, iii, ptr, ta;
   double probs[PNG], *freqs;
   int   i, j, k, t, iscale, nch, nam;
   int   iphen[NP], ifou[NP];
   double flik[BIG];
   double sum, p1, p2, term, u;
   int   nf, nb;

   ta = loci[wloc].n_all;
   ngenty = (ta * (ta + 1)) / 2;
   for (j = 1; j <= ngenty; j++)
      for (k = 1; k <= ngenty; k++)
         pene_mat[j][k] = (j == k) ? (1 - eps) : eps / ((double) ngenty - 1);

   freqs = loci[wloc].freq;
   iscale = 0;

   for (i = 0; i < NP; ++i)
   {
      ifou[i] = 0;
      iphen[i] = 0;
      cp_fou[i] = 0;
      itru_geno[i] = 0;
   }

   npeph = pgif[1][fam_numb][0];
   for (j = 1; j <= npeph; ++j)
   {
      nam = pgif[3][fam_numb][j];
      iphen[nam] = pgif[2 - rp_flg][fam_numb][j];
      ifou[nam] = pgif[4][fam_numb][j];
      cp_fou[nam] = ifou[nam];
   }

   for (i = 0; i < NST; ++i)
   {
      for (j = 0; j < NI; ++j)
         namsto[i][j] = 0;
      ndex[i] = 0;
   }

   nstore = 1;
   nstart[1] = 0;
   nstart[2] = ngenty;
   namsto[1][1] = NP + 1;
   ndex[1] = 1;
   nindi[1] = 1;
   k = 0;
   for (i = 1; i <= ta; i++)
      for (j = i; j <= ta; j++)
      {
         k++;
         if (i == j)
            store[k] = freqs[i - 1] * freqs[i - 1];
         else
            store[k] = 2 * freqs[i - 1] * freqs[j - 1];
      }
   sum = 0.0;
   for (j = 1; j <= ngenty; ++j)
      sum += store[j];
   if ((sum > 1.00001) || (sum < 0.99999))
   {
      printf ("\n Exited in geno freqs summation\n");
      exit (1);
   }
   noutin = 1;
   ptr = 1;
   while (noutin > 0)                   /* BEGIN PEELING; initialize cluster-like variables */
   {
      noutin = seq[ptr];
      ptr++;
      for (i = 1; i <= noutin; ++i)
      {
         invol[i] = seq[ptr];
         ptr++;
      }
      fflush (stdout);
      ntrip = 0;
      npar = 0;
      ninvol = noutin;
      nsib = seq[ptr];
      ptr++;
      nlik = seq[ptr];
      ptr++;
      iwrt = seq[ptr];
      ptr++;

      for (i = 1; i <= nsib; ++i)
      {
         im = seq[ptr];
         ptr++;
         ife = seq[ptr];
         ptr++;
         nch = seq[ptr];
         ptr++;

         upp_count++;
         upp[upp_count][1] = im;
         upp[upp_count][2] = ife;

         for (j = 1; j <= nch; ++j)
         {
            ich[j] = seq[ptr];
            ptr++;
         }
         ifoum = ifou[im];
         ifou[im] = 0;
         ifoufe = ifou[ife];
         ifou[ife] = 0;

         j = 1;
         while ((im != invol[j]) && (j <= ninvol))
            ++j;
         iim = j;
         if (j == ninvol + 1)
         {
            ++ninvol;
            invol[j] = im;
         }

         j = 1;
         while ((ife != invol[j]) && (j <= ninvol))
            ++j;
         iife = j;
         if (j == ninvol + 1)
         {
            ++ninvol;
            invol[j] = ife;
         }

         for (k = 1; k <= nch; ++k)
         {
            j = 1;
            while ((ich[k] != invol[j]) && (j <= ninvol))
               ++j;
            if (j == ninvol + 1)
            {
               ++ninvol;
               invol[j] = ich[k];
            }

            ++ntrip;
            itrip[1][ntrip] = iim;
            itrip[2][ntrip] = iife;
            itrip[3][ntrip] = j;
            mtrip[ntrip] = iim;

            if (iife > iim)
               mtrip[ntrip] = iife;
            if (j > mtrip[ntrip])
               mtrip[ntrip] = j;
         }

         /* Pop. geno frequencies for founders, considered ONCE and ONLY ONCE */
         if (ifoufe == 1)
         {
            ++npar;
            nprlik[npar] = 1;
            iprlik[1][npar] = iife;
            maxlik[npar] = iife;
            nstlik[npar] = 0;
         }
         if (ifoum == 1)
         {
            ++npar;
            nprlik[npar] = 1;
            iprlik[1][npar] = iim;
            maxlik[npar] = iim;
            nstlik[npar] = 0;
         }
      }
      nlik += npar;

      prev_functions (npar + 1, nlik, seq, &ptr);

      /* Efficient code for peeling starts here!!   */
      if (!noutin)                      /* The peeling operation involves a singleton */
      {
         /* In this loop, the singleton's id is stored in invol[1] */
         flik[1] = 0.0;
         for (i = 1; i <= ngenty; i++)  /* Calculate the overall likelihood */
         {
            term = pene_mat[i][iphen[invol[1]]];
            term *= store[nstlik[1] + i]; /* the first R-func listed is the     */
            /* only R-func, and there aren't      */
            /* any founders                       */
            flik[1] += term;
         }
      }                                 /* ends the if on peeling operation involves a singleton */

      else                              /* The peeling operation involves a cluster */
      {
         /* In this loop, the pivot's id is stored in invol[1] */

         if (invol[1] == im || invol[1] == ife) /*  pivot is a parent  */
         {
            ii = 0;
            jjj = npar + 1;
            while ((jjj <= nlik) && (ii == 0))
            {
               if (iprlik[1][jjj] == 1)
                  ii = 1;               /* pivot has prev R-func */
               jjj++;
            }
            if (ii)
               ii = jjj - 1;

            for (i = 1; i <= ngenty; i++) /* i indexes the pivot's possible genos */
            {
               flik[i] = 0.0;
               for (j = 1; j <= ngenty; j++) /* sum over nonpivot parent's genos  */
               {
                  term = 1.0;
                  for (k = 1; k <= ntrip; k++) /* for each kid, get contribution */
                  {
                     p2 = 0.0;
                     kkk = itrip[3][k];
                     jj = 0;
                     jjj = npar + 1;
                     while ((jjj <= nlik) && (jj == 0))
                     {
                        if (kkk == iprlik[1][jjj])
                           jj = 1;      /* kid has prev R-func */
                        jjj++;
                     }
                     for (kk = 1; kk <= ngenty; kk++) /* sum over kkk's genos   */
                     {
                        p1 = 1.0;
                        p1 *= sege_mat[i][j][kk];
                        p1 *= pene_mat[kk][iphen[invol[kkk]]];
                        if (jj)
                           p1 *= store[nstlik[jjj - 1] + kk];
                        p2 += p1;
                     }
                     term *= p2;
                     if (term == 0)
                        break;
                  }

                  /* Include pivot's prev R-func, which (if present) 
                     automatically incorporates population geno frequencies       */
                  if (ii > 0)
                     term *= store[nstlik[ii] + i];

                  /* If pivot does not have a prev R-func AND the pivot is a 
                     founder, then pop geno freqs are needed                     */
                  if (ii == 0 && cp_fou[invol[1]])
                     term *= store[i];

                  /* If pivot does not have a prev R-func and is NOT a founder, 
                     then the pivot's remaining contribution to the lkhd only 
                     occurs through his/her segregations, which have already 
                     been scored                                                 */

                  /* Either the nonpivot parent has prev R-func or his/her pop 
                     geno freqs are needed                                       */
                  jj = 0;
                  jjj = npar + 1;
                  while ((jjj <= nlik) && (jj == 0))
                  {
                     if (iprlik[1][jjj] == 2)
                        jj = 1;         /* nonpivot parent prev R-func */
                     jjj++;
                  }
                  if (jj)
                     term *= (store[nstlik[jjj - 1] + j] *
                              pene_mat[j][iphen[invol[2]]]);
                  else
                     term *= (store[j] * pene_mat[j][iphen[invol[2]]]);
                  flik[i] += term;
               }                        /* ends the loop on nonpivot parent's genos */
            }                           /* ends the loop on pivot's genos */
         }                              /* ends the IF on pivot's status, which is PARENT */

         else                           /* pivot is a kid */
         {
            /* find location of pivot's prev R-func */
            iii = 0;
            jjj = npar + 1;
            while ((jjj <= nlik) && (iii == 0))
            {
               if (iprlik[1][jjj] == 1)
                  iii = 1;              /* pivot has prev R-func */
               jjj++;
            }
            if (iii)
               iii = jjj - 1;

            /* find location of first parent's prev R-func */
            par1 = 0;
            jjj = npar + 1;
            while ((jjj <= nlik) && (par1 == 0))
            {
               if (iprlik[1][jjj] == 2)
                  par1 = 1;             /* par1 has prev R-func */
               jjj++;
            }
            if (par1)
               par1 = jjj - 1;

            /* find location of second parent's prev R-func */
            par2 = 0;
            jjj = npar + 1;
            while ((jjj <= nlik) && (par2 == 0))
            {
               if (iprlik[1][jjj] == 3)
                  par2 = 1;             /* par2 has prev R-func */
               jjj++;
            }
            if (par2)
               par2 = jjj - 1;

            for (i = 1; i <= ngenty; i++) /* i indexes the genos of the pivot */
            {
               flik[i] = 0.0;
               for (j = 1; j <= ngenty; j++) /* sum genos of par1 (j) and par2 (jj) */
               {
                  for (jj = 1; jj <= ngenty; jj++)
                  {
                     term = 1.0;
                     for (k = 1; k <= ntrip; k++) /* get each kids contribution */
                     {                  /* except for the pivot       */
                        p2 = 0.0;
                        kkk = itrip[3][k];
                        if (kkk != 1)
                        {
                           ii = 0;
                           jjj = npar + 1;
                           while ((jjj <= nlik) && (ii == 0))
                           {
                              if (kkk == iprlik[1][jjj])
                                 ii = 1; /* kid's prev R-func */
                              jjj++;
                           }
                           for (kk = 1; kk <= ngenty; kk++) /* sum over kkk's genos */
                           {
                              p1 = 1.0;
                              p1 *= sege_mat[j][jj][kk];
                              p1 *= pene_mat[kk][iphen[invol[kkk]]];
                              if (ii)
                                 p1 *= store[nstlik[jjj - 1] + kk];
                              p2 += p1;
                           }
                           term *= p2;
                           if (term == 0)
                              break;
                        }               /* ends IF on pivot exclusion */
                     }                  /* end FOR loop on sibship */

                     term *= sege_mat[j][jj][i]; /* pivot segs and */
                     if (iii > 0)
                        term *= store[nstlik[iii] + i]; /* prev R-func    */
                     term *= pene_mat[j][iphen[invol[2]]]; /* insert parents */
                     term *= pene_mat[jj][iphen[invol[3]]]; /* phenos         */

                     /* If par1 has no prev R-func then pop geno freqs are needed  */
                     if (par1 == 0)
                        term *= store[j];
                     else
                        term *= store[nstlik[par1] + j];

                     /* If par2 has no prev R-func then pop geno freqs are needed  */
                     if (par2 == 0)
                        term *= store[jj];
                     else
                        term *= store[nstlik[par2] + jj];

                     flik[i] += term;
                  }
               }                        /* end both FOR loops on parent-pair genos */
            }                           /* end loop on pivot's genos */
         }                              /* ends the else on pivot's status, which is KID */
      }                                 /* ends the else on peeling operation involving a cluster */

      /* Efficient code for peeling ends here */

      if (noutin > 0)
      {
         /* rescale the R-function if necessary ( and check non-zero)  */
         iscale += rescale (flik, ngenty);
         /* call REPACK if space exhausted */
         nf = nstart[nstore + 1] + ngenty;
         if ((nf > 2 * BIG) || (nstore >= NST - 2))
            repack ();

         /* store computed R-function in appropriate position */
         nf = nstart[nstore + 1] + ngenty;
         if ((nf > 2 * BIG) || (nstore >= NST - 2))
         {
            printf ("\n Exited because nstart too small\n");
            exit (1);
         }

         nstore += 1;
         nstart[nstore + 1] = nf;
         nb = nstart[nstore];
         for (j = 1; j <= ngenty; ++j)
            store[nb + j] = flik[j];
         ndex[nstore] = 1;
         nindi[nstore] = noutin;
         namsto[nstore][0] = ninvol;
         for (j = 1; j <= ninvol; j++)
            namsto[nstore][j] = invol[j];

         varlik (flik, ngenty);
      }                                 /* One cluster's contribution to the lkhd is stored */
   }                                    /* Ends the WHILE on noutin; the entire family has now been peeled */
   *sf = iscale;

   rlen = nstore;
   if (!rp_flg)
   {
      /* printf("Likelihood %e, scaled %d times \n",flik[1],iscale); */
      return (flik[1]);
   }

   for (nb = 1; nb <= CHAIN; nb++)      /* BEGIN REVERSE PEELING if rp_flg = 1 */
   {
      u = rans ();                      /* back sample the last pivot, first */
      k = (nstore - 1) * ngenty;
      pivot = namsto[nstore][1];
      i = 1;
      probs[0] = (store[k + i] * pene_mat[i][iphen[pivot]]) / flik[1];
      while (u >= probs[0] && probs[0] < 1)
      {
         i++;
         probs[0] += (store[k + i] * pene_mat[i][iphen[pivot]]) / flik[1];
      }
      itru_geno[pivot] = i;

      while (nstore > 1)                /* traverse the peeling sequence in reverse order    */
      {
         t = namsto[nstore][0];
         for (i = 0; i < 4; i++)
            for (j = 0; j <= t; j++)
               cluster[i][j] = 0;       /* init cluster */
         cluster[0][0] = t;             /* set cluster size */
         pivot = namsto[nstore][1];     /* store pivot's id */

         for (i = 1; i <= cluster[0][0]; i++)
         {
            cluster[0][i] = namsto[nstore][i]; /* id's of the cluster members */
            cluster[1][i] = cp_fou[cluster[0][i]]; /* and their fou. statuses */
            j = nstore;
            while (j > 1 && cluster[0][i] != namsto[j][1])
               j--;                     /* last prev_rf (if any) */
            if (j > 1 && cluster[0][i] != pivot)
               cluster[2][i] = (j - 1) * ngenty;
            else
               cluster[2][i] = 0;
         }

         /* Determine whether parents are in positions 2 and 3, or 1 and 2.   */

         i = 0;
         par1 = 0;
         while (i <= upp_count)
         {
            i++;
            if (cluster[0][1] == upp[i][1] && cluster[0][2] == upp[i][2])
               par1 = 1;
            if (cluster[0][1] == upp[i][2] && cluster[0][2] == upp[i][1])
               par1 = 1;
            if (cluster[0][2] == upp[i][1] && cluster[0][3] == upp[i][2])
               par1 = 2;
            if (cluster[0][2] == upp[i][2] && cluster[0][3] == upp[i][1])
               par1 = 2;
         }
         par2 = par1 + 1;               /*  found unique parent pair   */
         cluster[3][par1] = 1;
         cluster[3][par2] = 1;

         pp = cluster[3][1];            /* pp=1 means that pivot is a parent 
                                           relative to this cluster           */
         if (pp == 0)                   /* switch the first and third cols    */
         {
            for (i = 0; i < 4; i++)
            {
               cluster[i][t + 1] = cluster[i][1];
               cluster[i][1] = cluster[i][3];
               cluster[i][3] = cluster[i][t + 1];
            }
         }
         /* begin back sampling by cluster */
         if (pp)                        /* pivot is a parent              */
         {
            cend = cluster[0][0];
            par1 = cluster[0][1];
            par2 = cluster[0][2];
            for (jj = 1; jj <= ngenty; jj++)
            {
               term = 1.0;
               j = itru_geno[par1];
               for (kk = 3; kk <= cend; kk++)
               {
                  p1 = 0;
                  off = cluster[0][kk];
                  kkk = cluster[2][kk];
                  if (kkk == 0)         /* off has no prev_rf */
                     for (k = 1; k <= ngenty; k++)
                        p1 += (pene_mat[k][iphen[off]] * sege_mat[j][jj][k]);
                  else                  /* off has prev_rf */
                     for (k = 1; k <= ngenty; k++)
                        p1 += (store[kkk + k] * pene_mat[k][iphen[off]]
                               * sege_mat[j][jj][k]);
                  term *= p1;
                  if (term == 0)
                     break;
               }
               if (term != 0)
               {
                  jjj = cluster[2][2];
                  if (cluster[1][2] == 1 && jjj == 0) /* founder w/o prev_rf */
                     term *= (pene_mat[jj][iphen[par2]] * store[jj]);
                  else
                     term *= (pene_mat[jj][iphen[par2]] * store[jjj + jj]);
               }
               probs[jj] = term;
            }

            u = rans ();
            sum = 0;                    /* back sample the other parent */
            for (k = 1; k <= ngenty; k++)
               sum += probs[k];
            k = 1;
            probs[0] = probs[1] / sum;
            while (u >= probs[0] && probs[0] < 1)
            {
               k++;
               probs[0] += (probs[k] / sum);
            }
            itru_geno[par2] = k;

            jj = itru_geno[par2];       /* back sample kids */
            for (kk = 3; kk <= cend; kk++)
            {
               off = cluster[0][kk];
               kkk = cluster[2][kk];
               if (kkk != 0)
                  for (k = 1; k <= ngenty; k++)
                     probs[k] =
                        store[kkk + k] * pene_mat[k][iphen[off]]
                        * sege_mat[j][jj][k];
               else
                  for (k = 1; k <= ngenty; k++)
                     probs[k] = pene_mat[k][iphen[off]] * sege_mat[j][jj][k];
               sum = 0;
               for (i = 1; i <= ngenty; i++)
                  sum += probs[i];
               u = rans ();
               i = 1;
               probs[0] = probs[i] / sum;
               while (u >= probs[0] && probs[0] < 1)
               {
                  i++;
                  probs[0] += (probs[i] / sum);
               }
               itru_geno[off] = i;
            }
         }
         else                           /* if pivot is not a parent  */
         {
            j = 0;
            jjj = itru_geno[pivot];
            cend = cluster[0][0];
            par1 = cluster[0][1];
            par2 = cluster[0][2];
            for (i = 1; i <= (ngenty * ngenty); i++) /* big loop on probs */
            {
               if (((i - 1) % ngenty) == 0)
               {
                  j++;
                  jj = 0;
               }
               jj++;
               term = 1.0;
               for (kk = 4; kk <= cend; kk++) /* cycle thru the kids */
               {
                  p1 = 0;
                  off = cluster[0][kk];
                  kkk = cluster[2][kk];
                  if (kkk == 0)         /* no prev_rf */
                     for (k = 1; k <= ngenty; k++)
                        p1 += (pene_mat[k][iphen[off]] * sege_mat[j][jj][k]);
                  else                  /* prev_rf */
                     for (k = 1; k <= ngenty; k++)
                        p1 += (store[kkk + k] * pene_mat[k][iphen[off]]
                               * sege_mat[j][jj][k]);
                  term *= p1;
                  if (term == 0)
                     break;             /* stop cycling thru kids */
               }
               if (term != 0)           /* cycle thru parents */
               {
                  term *= sege_mat[j][jj][jjj];
                  k = j;
                  for (kk = 1; kk <= 2; kk++)
                  {
                     ii = cluster[0][kk];
                     kkk = cluster[2][kk];
                     if (cluster[1][kk] == 1 && kkk == 0) /* if founder and no prev-rf */
                        term *= (store[k] * pene_mat[k][iphen[ii]]);
                     else               /* nonfounder with prev_rf */
                        term *= (store[kkk + k] * pene_mat[k][iphen[ii]]);
                     k = jj;
                  }
               }
               probs[i] = term;
            }                           /* done with conditional probs for parent pair */

            sum = 0;
            for (i = 1; i <= (ngenty * ngenty); i++)
               sum += probs[i];
            u = rans ();
            i = 1;
            probs[0] = probs[i] / sum;
            while (u >= probs[0] && probs[0] < 1)
            {
               i++;
               probs[0] += (probs[i] / sum);
            }
            if ((i % ngenty) != 0)
               itru_geno[par2] = i % ngenty;
            else
               itru_geno[par2] = ngenty;
            itru_geno[par1] = ((i - itru_geno[par2]) / ngenty + 1); /* parents sampled */

            j = itru_geno[par1];
            jj = itru_geno[par2];       /* back sample kids */
            for (kk = 4; kk <= cend; kk++)
            {
               off = cluster[0][kk];
               kkk = cluster[2][kk];
               if (kkk != 0)
                  for (k = 1; k <= ngenty; k++)
                     probs[k] =
                        store[kkk + k] * pene_mat[k][iphen[off]]
                        * sege_mat[j][jj][k];
               else
                  for (k = 1; k <= ngenty; k++)
                     probs[k] = pene_mat[k][iphen[off]] * sege_mat[j][jj][k];
               sum = 0;
               for (i = 1; i <= ngenty; i++)
                  sum += probs[i];
               u = rans ();
               i = 1;
               probs[0] = probs[i] / sum;
               while (u >= probs[0] && probs[0] < 1)
               {
                  i++;
                  probs[0] += (probs[i] / sum);
               }
               itru_geno[off] = i;
            }
         }
         nstore--;
      }

      for (i = 1; i <= npeph; i++)
      {
         j = pgif[3][fam_numb][i];
         if (pgif[1][fam_numb][i])
            pgif[2][fam_numb][i] = itru_geno[j];
      }

      nstore = rlen;                    /* resets nstore for another round of reverse peeling    */
   }
   return (1.0);
}

/* Notes: 
          
      (i) A cluster is a nuclear subfamily that is connected to the rest of the
          family through a single pivot. A singleton cannot be a cluster. The 
	  reverse-peeling portion of this progam assumes that all families are 
	  zero-looped families, which implies that all families can be 
	  decomposed into a series of conditionally independent clusters given 
	  the complete set of pivots.
 
     (ii) The variable 'cluster' is a matrix of dimension (4 x 'cluster_size').
          The columns of 'cluster' are one-indexed, the rows are not.

    (iii) 0th-row contains cluster member ids, w/ parents followed by kids 
          1st-row contains the founder status of ea. member
          2nd-row contains the positions of stored R-functs and is 0 if no 
	          R-funct exists
          3rd-row contains an indicator of which members are the parents

     (iv) The pivot for any cluster is in position 1, 2, or 3.

      (v) The penetrance terms of pivots, are NOT included in any R-funct. 
          The penetrance of the final singleton is incorporated into the 
	  likelihood calculation when the final operation of the peeling 
	  sequence (e.g. "0 0 1 2 0 0") is executed. In particular, the terms
	  which are summed to produce the overall likelihood are not stored 
	  anywhere, except in the trival case that the singleton does not 
	  provide data (i.e. his/her penetrance terms are all 1.0).

     (vi) If a member of the cluster is a founder, his/her genotype frequencies 
          ARE included in the R-function for the pivot of that cluster. This 
	  is accomplished through the use of prev_functions() which grabs 
	  additional R-functs as needed, and in the case of founders, the 
	  additional R-funct corresponds to population genotype frequencies. 
	  At this time, his/her founder status is toggled, so 
	  that the founder genotype frequencies are not used in any subsequent 
	  calculation (e.g. additional marriages or singleton calculations). 
	  This is why, reverse-peeling requires a copy of the original binary 
	  vector indicating the "true" founder status of all individuals.     */
