/*
 * civil_readpar.c
 *
 *  Created on: Jul 14, 2009
 *      Author: yanming
 */

/**
 * Skip the rest of the current line.
 *
 * @param fp pointer to the input file
 */
void next_line (FILE * fp)
{
   char  dummy[MAX_LEN];
   fgets (dummy, MAX_LEN, fp);
}

/**
 * Compare names of two individuals. Called by extract_ped().
 * @param pt1, pt2 two individuals.
 */
int name_cmp (const void *pt1, const void *pt2)
{
   return strcmp (((name_id_t *) pt1)->name, ((name_id_t *) pt2)->name);
}

/**
 * Extract a sub-pedigree from a super-pedigree.  Extracts the individuals
 * indexed by i1, ..., i2 from the super-pedigree, stores them in a
 * sub-pedigree, and relabels them to 0, 1, ... in the sub-pedigree.
 *
 * We use the term pedigree in a broad sense. Not any two people in the
 * pedigree have to be related. Both the super-pedigree and sub-pedigree can
 * have more than one connected components. For example, the super-pedigree
 * can be the collection of all individuals in the data. The sub-pedigree can
 * be the individuals in one connected component (a pedigree in a narrow
 * sense).
 *
 * NOTE: In the super-pedigree, the individuals need to be in chronological
 * order. Parents are listed before their children.
 *
 * @param superped_pt a pointer to the super-pedigree.
 *
 * @param i1, i2 the starting and ending indices to the individuals in the
 * super-pedigree that we want to extract. So we can only extract consecutive
 * individuals from the super-pedigree.
 *
 * @param ped_pt pointer to the sub-pedigree.
 *
 */
void extract_ped (ped_t * superped_pt, int i1, int i2, ped_t * ped_pt)
{

   name_id_t *name_id_list, target, *result;
   /* We use these for matching names to corresponding ids. */

   int   i;                             /* indexes individuals */

   sprintf (ped_pt->name, "%s_%d_%d", superped_pt->name, i1, i2);
   ped_pt->size = i2 - i1 + 1;
   ped_pt->inds = (ind_t *) malloc (sizeof (ind_t) * ped_pt->size);
   name_id_list = (name_id_t *) malloc (sizeof (name_id_t) * ped_pt->size);

   /* Copy individuals from the super-pedigree to the sub-pedigree. */
   memcpy (ped_pt->inds, &superped_pt->inds[i1], sizeof (ind_t) * ped_pt->size);

   /* Relabel individuals in the sub-pedigree. Build a list of all (id, name)
    */
   /* pairs, and sort the list by name. */
   for (i = 0; i < ped_pt->size; i++)
   {
      ped_pt->inds[i].id = i;
      name_id_list[i].id = i;
      strcpy (name_id_list[i].name, ped_pt->inds[i].name);
   }
   qsort (name_id_list, ped_pt->size, sizeof (name_id_t), name_cmp);

   /* For each individuals, find his/her mother's id and father's id, by */
   /* matching the names. For this part to work, parents need to be listed */
   /* before their children. */
   for (i = 0; i < ped_pt->size; i++)
   {
      /* Look up father's id. */
      strcpy (target.name, ped_pt->inds[i].pa_name);
      result = bsearch (&target, name_id_list, ped_pt->size,
                        sizeof (name_id_t), name_cmp);
      if (result == NULL)
      {
         ped_pt->inds[i].pa_id = -1;
      }
      else
      {
         ped_pt->inds[i].pa_id = result->id;
      }

      /* Look up mother's id. */
      strcpy (target.name, ped_pt->inds[i].ma_name);
      result = bsearch (&target, name_id_list, ped_pt->size,
                        sizeof (name_id_t), name_cmp);
      if (result == NULL)
      {
         ped_pt->inds[i].ma_id = -1;
      }
      else
      {
         ped_pt->inds[i].ma_id = result->id;
      }
   }

   /* print_ped(ped_pt); */
   free (name_id_list);
}

/**
 * Populate my data structures with MORGAN global variables.
 *
 * @param name_ped: name/key word of the data set. Will use this to name
 * output files.
 *
 * @param ped_list_pt: points to a list of pedigrees.
 *
 */
void read_peds_MORGAN (char *name_ped, ped_list_t * ped_list_pt)
{

   ped_t superped;                      /* stores individuals from all components in a single pedigree. */
   int   n_peds;                        /* number of pedigrees. */
   int  *ped_starts, *ped_ends;         /* indices to individuals starting/ending a component. */

   int   i;                             /*indexes individuals */
   int   k;                             /*indexes pedigrees */

   double sy, sy2;                      /* sum of the trait values, sum of the squared trait values */
   int   ny;                            /* number of observed individuals */

   Ind  *mum, *dad;

   /* pedigree size */
   superped.size = pedsize;
   fprintf (stdout, "Total number of individuals in the pedigree file: %d\n",
            superped.size);

   /* pedigree name */
   strcpy (superped.name, name_ped);

   superped.inds = (ind_t *) malloc (sizeof (ind_t) * superped.size);
   /* fprintf(stdmsg, "Done allocating memory. \n"); */

   for (i = 0; i < superped.size; i++)
   {
/*            fprintf(stdout, "%d ", i); */

      superped.inds[i].id = nghd[i].index;
      strcpy (superped.inds[i].name, nghd[i].self);

      /* Indices to component, patmeio, matmeio start from 1 in MORGAN */
      superped.inds[i].cmp = nghd[i].component - 1;
      superped.inds[i].pa_mi = nghd[i].patmeio - 1;
      superped.inds[i].ma_mi = nghd[i].matmeio - 1;

      /* superped.inds[i].pa_name, superped.inds[i].ma_name */
      if (!isNullNode (nghd[i].pa))
      {
         dad = nghd[i].pa;
         mum = nghd[i].ma;
         strcpy (superped.inds[i].pa_name, nghd[dad->index].self);
         strcpy (superped.inds[i].ma_name, nghd[mum->index].self);
         superped.inds[i].pa_id = dad->index;
         superped.inds[i].ma_id = mum->index;
      }
      else
      {
         superped.inds[i].pa_id = -1;
         superped.inds[i].ma_id = -1;
      }

      /* trait value */
      superped.inds[i].obsd = nghd[i].obs;
      if (nghd[i].obs)
      {
         superped.inds[i].y = nghd[i].q_dat[0];
      }
   }

/*      printf("\n"); */

/*      print_ped(&superped); */

   /* compute the sample mean and sd of the trait values */
   sy = sy2 = 0;
   ny = 0;
   for (i = 0; i < superped.size; i++)
   {
      if (superped.inds[i].obsd)
      {
         ny++;
         sy += superped.inds[i].y;
         sy2 += superped.inds[i].y * superped.inds[i].y;
      }
   }

   ped_list_pt->mu = sy / ny;
   ped_list_pt->sd = sqrt ((sy2 - sy * sy / ny) / (ny - 1));

   /* TODO:
    *  To split/merge pedigrees, we can modify the pedigree component id here.
    */
   /* -- Split <superped> into a list of subpedigrees (by component). */

   fprintf (stdmsg, "Identifying pedigree components ...\n");
   /* Count the total number of pedigrees(components). */
   n_peds = 0;
   for (i = 1; i < superped.size; i++)
   {
      if (superped.inds[i].cmp != superped.inds[i - 1].cmp)
      {
         n_peds++;
      }
   }
   n_peds++;
   fprintf (stdout, "There are %d pedigree components. \n", n_peds);

   /* Find the indices to individuals that start/end each component. */
   ped_starts = (int *) malloc (sizeof (int) * n_peds);
   ped_ends = (int *) malloc (sizeof (int) * n_peds);
   ped_starts[0] = 0;
   k = 0;
   for (i = 1; i < superped.size; i++)
   {
      if (superped.inds[i].cmp != superped.inds[i - 1].cmp)
      {
         ped_ends[k] = i - 1;
         k++;
         ped_starts[k] = i;
      }
   }
   ped_ends[k] = superped.size - 1;

   /* Split <superped> into a list of pedigrees (by component) */
   ped_list_pt->n_peds = n_peds;
   ped_list_pt->peds = (ped_t *) malloc (sizeof (ped_t) * n_peds);

   /* fprintf(stdout, "%8s  %8s\n", "ped no.", "size"); */
   for (k = 0; k < n_peds; k++)
   {
      extract_ped (&superped, ped_starts[k], ped_ends[k],
                   &ped_list_pt->peds[k]);
      /* fprintf(stdout, "%8d, %8d \n", k, ped_list_pt->peds[k].size); */
   }
   fprintf (stdout, "\n");

   /*
      fprintf(stdmsg, "Free unused memories. \n");
    */
   free_pedi (&superped);
   free (ped_starts);
   free (ped_ends);

}

/**
 * When the lines in the inheritance vector file are arranged by component, we need to rearrange the them so that they will be ordered by MCMC run.
 * The rearranged lines will be written into a new file. The name of the new
 * file is the name of the old meiosis indicator file plus the suffix ".rearranged".
 *
 * WARNING: the program will not check whether a file with the same name exists. If it does, it will be overwritten.
 *
 * @param dat_pars_pt points to date set parameters
 *
 * @param fns_pt points to file names (we need the name of the
 *
 */
void rearrange_iv_file (data_pars_t * dat_pars_pt, file_names_t * fns_pt)
{
   FILE *fp_in, *fp_out;
   char  dummy[MAX_LEN];

   long  nlines;                        /* number of lines in the input file */
   long  i;                             /* indexes meioses */
   int   k;                             /* indexes components */
   int   h;                             /* indexes the MCMC runs */
   long *lns;                           /* indexes line numbers, corresponding different components */
   fpos_t *pos;                         /* starting positions of the lines in the input file */

   /* open the original meiosis indicator file */
   if ((fp_in = fopen (fns_pt->file_mis, "r")) == NULL)
   {
      printf ("Cannot open file (meiosis indicators) %s for reading. \n",
              fns_pt->file_mis);
      exit (1);
   }

   /* Build an index of line starting positions */
   nlines = (long) dat_pars_pt->n_meis * dat_pars_pt->smpl_size;

   /*
      fprintf(stdmsg, "nlines: %d x %d = %ld\n",
      dat_pars_pt->n_meis, dat_pars_pt->smpl_size, nlines);
    */

   pos = (fpos_t *) malloc (sizeof (fpos_t) * nlines);
   for (i = 0; i < nlines; i++)
   {
      if (fgetpos (fp_in, &pos[i]))
      {
         perror ("fgetpos failed");
         exit (1);
      };
      fgets (dummy, MAX_LEN, fp_in);
   }

   fprintf (stdmsg, "Done building the index of line starting positions.\n");
   fflush (stdmsg);

   /* create a new file to store the rearranged lines */
   strcpy (fns_pt->file_mis_rearranged, fns_pt->file_mis);
   strcat (fns_pt->file_mis_rearranged, ".rearranged");
   if ((fp_out = fopen (fns_pt->file_mis_rearranged, "w")) == NULL)
   {
      printf ("Cannot open file (meiosis indicators) %s for reading. \n",
              fns_pt->file_mis);
      exit (1);
   }

   /* In the input file, the lines are ordered as follows:
    *     MCMC block 1 for component 1
    *     MCMC block 2 for component 1
    *     ...
    *     MCMC block H for component 1
    *     MCMC block 1 for component 2
    *     MCMC block 2 for component 2
    *     ...
    *     MCMC block H for component 2
    *     ...
    * We need to rearrange the lines, so that they are ordered as follows:
    *     MCMC block 1 for component 1
    *     MCMC block 1 for component 2
    *     ...
    *     MCMC block 1 for component K
    *     MCMC block 2 for component 1
    *     MCMC block 2 for component 2
    *     ...
    *     MCMC block 2 for component K
    *     ...
    *
    * After we have read one MCMC block from all components, we go back to the first component and read the second MCMC block.
    */

   lns = (long *) malloc (sizeof (long) * dat_pars_pt->n_components);

   /* Compute the starting line numbers for each pedigree component in the input mis file */
   lns[0] = 0;
   for (k = 1; k < dat_pars_pt->n_components; k++)
   {
      lns[k] = lns[k - 1] + dat_pars_pt->n_meis_ped[k - 1]
         * dat_pars_pt->smpl_size;
   }

   for (h = 0; h < dat_pars_pt->smpl_size; h++)
   {
      for (k = 0; k < dat_pars_pt->n_components; k++)
      {
         /* Read one MCMC block of component k and write the block to the output file */
         for (i = 0; i < dat_pars_pt->n_meis_ped[k]; i++)
         {
            printf ("%ld\n", lns[k]);
            if (fsetpos (fp_in, &pos[lns[k]]))
            {
               perror ("fsetpos failed");
               exit (1);
            }
            else
            {
               /* fgets(dummy, line_max, fp); */
               /* printf("%s", dummy); */
               dummy[0] = 0;
               fscanf (fp_in, "%[^\n]", dummy);
               fprintf (fp_out, "%s\n", dummy);
               lns[k] += 1;
            }
         }
      }

   }

   fprintf (stdmsg, "Done rearranging the lines. \n");
   fflush (stdmsg);
   free (pos);
   free (lns);
   fclose (fp_in);
   fclose (fp_out);
}

/**
 * Read in the meiosis indicator matrix and convert the meiosis indicator
 * matrix to inheritance vectors. In the input file, each row corresponds to
 * one meiosis in one data set. Within each row, the first entry is the
 * meiosis indicator at a unlinked locus. The rest of the row encodes meiosis
 * indicators along a chromosome (marker map). The second entry specifies the
 * meiosis indicator at the first locus of a marker map.  The third entry
 * specifies how many switches(recombinations) are there in this meiosis. The
 * rest of the entries list all the positions at which recombination (switch)
 * happens. (This is equivalent to a run-length encoding.)
 *
 * @param file_mis name of the input file.
 *
 * @param dat_pars_pt points to data set parameters
 *
 * @psram adg_list_pt points to a list of pedigrees (allele descent graphs).
 *
 * @param iv_list_pt stores the resulted inheritance vectors, of dimensions smpl_size x n_loci x n_peds.
 *
 * @param iv_unlinked_pt stores inheritance vectors at the unlinked locus, of dimensions smpl_size x 1 x n_peds.
 */

void read_ivs (char *file_mis, const data_pars_t * dat_pars_pt,
               const adg_list_t * adg_list_pt, iv_list_t * iv_list_pt,
               iv_list_t * iv_unlinked_pt)
{

   int   smpl_size;
   int   n_meis;
   int   n_loci, n_pos;
   char  dummy[MAX_LEN];

   FILE *fp;
   int   h;                             /*ndexes sample points */
   int   k;                             /*indexes pedigrees */
   int   i;                             /*indexes meioses */
   int   i_mi;                          /*indexes meioses indicators */
   int   j, j0, j1, j2;                 /*index loci */
   int   b0, b1;                        /* meiosis indicators */
   int   n;                             /* number of switches */

   char *miv, **mi_mat;

   char  bit;

   typedef struct
   {
      char  first;                      /* meiosis indicator at the first locus */
      int   n_swis;                     /* number of switches */
      int  *swis;                       /* switch positions */
   } mic_t;                             /* a chain of meiosis indicators along a chromosome (same meiosis, different loci) */

   mic_t *mics;                         /* meiosis indicator chains at all meioses */

   smpl_size = dat_pars_pt->smpl_size;
   n_meis = dat_pars_pt->n_meis;
   n_loci = dat_pars_pt->n_loci;
   n_pos = dat_pars_pt->n_pos;

   /*  Allocate memories */
   fprintf (stdmsg, "Allocating memories ...\n");
   mics = (mic_t *) malloc (sizeof (mic_t) * n_meis);
   miv = (char *) malloc (sizeof (char) * n_meis);
   mi_mat = (char **) malloc (sizeof (char *) * n_meis);

   /*  We allocate memories ONLY for possible test positions! */
   iv_list_pt->smpl_size = smpl_size;
   iv_list_pt->n_loci = n_loci;
   iv_list_pt->n_peds = adg_list_pt->n_peds;
   iv_list_pt->s = (iv_t ***) malloc (sizeof (iv_t **) * iv_list_pt->smpl_size);
   for (h = 0; h < iv_list_pt->smpl_size; h++)
   {
      iv_list_pt->s[h] = (iv_t **) malloc (sizeof (iv_t *)
                                           * iv_list_pt->n_loci);
      for (j = 0; j < n_loci; j++)
      {
         iv_list_pt->s[h][j] = NULL;
      }
      for (j = 0; j < n_pos; j++)
      {
         iv_list_pt->s[h][dat_pars_pt->pos[j]] =
            (iv_t *) malloc (sizeof (iv_t) * iv_list_pt->n_peds);
      }
   }

   iv_unlinked_pt->smpl_size = smpl_size;
   iv_unlinked_pt->n_loci = 1;
   iv_unlinked_pt->n_peds = adg_list_pt->n_peds;
   iv_unlinked_pt->s = (iv_t ***) malloc (sizeof (iv_t **)
                                          * iv_unlinked_pt->smpl_size);
   for (h = 0; h < iv_unlinked_pt->smpl_size; h++)
   {
      iv_unlinked_pt->s[h] = (iv_t **) malloc (sizeof (iv_t *)
                                               * iv_unlinked_pt->n_loci);
      for (j = 0; j < iv_unlinked_pt->n_loci; j++)
      {
         iv_unlinked_pt->s[h][j] = (iv_t *) malloc (sizeof (iv_t)
                                                    * iv_unlinked_pt->n_peds);
      }
   }

   for (i = 0; i < n_meis; i++)
   {
      mi_mat[i] = (char *) malloc (sizeof (char) * n_loci);
   }

   fprintf (stdmsg, "Done allocating memories ...\n");

   if ((fp = fopen (file_mis, "r")) == NULL)
   {
      printf ("Cannot open file (meiosis indicators) %s for reading. \n",
              file_mis);
      exit (1);
   }

   for (h = 0; h < smpl_size; h++)
   {

      /*  Read one set of meiosis indicator chains */
      for (i = 0; i < n_meis; i++)
      {
         fscanf (fp, "%s", dummy);
         fscanf (fp, "%d %d %d", &b0, &b1, &n);
         miv[i] = (char) b0;

         mics[i].first = (char) b1;
         mics[i].n_swis = n;
         if (n > 0)
         {
            mics[i].swis = (int *) malloc (sizeof (int) * n);
         }
         else
         {
            mics[i].swis = NULL;
         }
         for (j = 0; j < n; j++)
         {
            fscanf (fp, "%d", &b1);
            mics[i].swis[j] = b1 - 1;
            /*  Minus the positions by 1 since array starts with 0 in C. */
         }
      }

      /*  Convert the meiosis indicator chains to a meiosis indicator matrix */
      for (i = 0; i < n_meis; i++)
      {
         j = 0;
         bit = mics[i].first;

         for (j1 = 0; j1 < mics[i].n_swis; j1++)
         {
            for (j2 = j; j2 < mics[i].swis[j1]; j2++)
            {
               mi_mat[i][j2] = bit;
            }
            bit = 1 - bit;
            j = j2;
         }

         /*  printf(": %d %d\n", j, n_loci); */
         for (j2 = j; j2 < n_loci; j2++)
         {
            mi_mat[i][j2] = bit;
         }
      }

      /*  Free mics[i].swis. We will re-allocate memory to them in the next */
      /*  loop. */
      for (i = 0; i < n_meis; i++)
      {
         free (mics[i].swis);
      }

      /*  Convert the meiosis indicator matrix to inheritance vectors */
      for (j0 = 0; j0 < n_pos; j0++)
      {
         j = dat_pars_pt->pos[j0];
         for (k = 0; k < adg_list_pt->n_peds; k++)
         {
            iv_list_pt->s[h][j][k] = 0;
            for (i = 0; i < adg_list_pt->peds[k].n_nfa; i++)
            {
               i_mi = adg_list_pt->peds[k].i_mi[i];
               if (i_mi >= 0)
               {                        /* some i_mi can be -1. */
                  iv_list_pt->s[h][j][k] += mi_mat[i_mi][j] << i;
               }
            }
         }
      }

      /*  Convert the meiosis vector to an inheritance vector */
      for (k = 0; k < adg_list_pt->n_peds; k++)
      {
         iv_unlinked_pt->s[h][0][k] = 0;
         for (i = 0; i < adg_list_pt->peds[k].n_nfa; i++)
         {
            i_mi = adg_list_pt->peds[k].i_mi[i];
            if (i_mi >= 0)
            {                           /* some i_mi can be -1. */
               iv_unlinked_pt->s[h][0][k] += miv[i_mi] << i;
            }
         }
      }

   }

   fclose (fp);

   /*  -- Locus map */
   iv_list_pt->map = (double *) malloc (sizeof (double) * iv_list_pt->n_loci);
   for (j = 0; j < n_loci; j++)
   {
      iv_list_pt->map[j] = dat_pars_pt->locus_map[j];
   }

   iv_unlinked_pt->map = (double *) malloc (sizeof (double)
                                            * iv_unlinked_pt->n_loci);
   iv_unlinked_pt->map[0] = 0;

   /*  Free memories */
   for (i = 0; i < n_meis; i++)
   {
      free (mi_mat[i]);
   }
   free (mi_mat);
   free (mics);
   free (miv);
}

/**
 * Read file names.
 *
 * @param fp points to the input file.
 *
 * @pt pt points to the data structure that stores file names.
 */
void read_file_names (FILE * fp, file_names_t * pt)
{
   /* Input file names */
   /*
      fscanf(fp, "%*[^\n]");
      fscanf(fp, "%s \n", pt->dir_in);
      fscanf(fp, "%*[^\n]");
      fscanf(fp, "%s \n", file_name);
      sprintf(pt->file_ped, "%s/%s", pt->dir_in, file_name);
      fscanf(fp, "%*[^\n]");
      fscanf(fp, "%s \n", file_name);
      sprintf(pt->file_trait, "%s/%s", pt->dir_in, file_name);
      fscanf(fp, "%*[^\n]");
      fscanf(fp, "%s \n", file_name);
      sprintf(pt->file_mis, "%s/%s", pt->dir_in, file_name);
    */

   /* Read the name of the meiosis indicator file */
   next_line (fp);

   fscanf (fp, "%s", pt->file_mis);
   next_line (fp);

   /* Output file names */
   next_line (fp);
   fscanf (fp, "%s", pt->dir_out);
   next_line (fp);

   next_line (fp);
   fscanf (fp, "%s", pt->name_ped);
   next_line (fp);
   /*
      sprintf(pt->file_miv_t, "%s/%s.miv.vc.t.out", pt->dir_out, pt->name_ped);
      sprintf(pt->file_miv_p, "%s/%s.miv.vc.p.out", pt->dir_out, pt->name_ped);
      sprintf(pt->file_civ_t, "%s/%s.civ.vc.t.out", pt->dir_out, pt->name_ped);
      sprintf(pt->file_civ_p, "%s/%s.civ.vc.p.out", pt->dir_out, pt->name_ped);
    */
}

/**
 * Read inheritance vector file info, including the number number of meioses,
 * number of MCMC runs, number of marker loci and marker map.
 *
 * @param fp points to the input file
 *
 * @param data_pars_pt points to the data structure storing the trait model
 * parameters.
 */
void read_data_pars (FILE * fp, data_pars_t * data_pars_pt)
{
   char  dummy[MAX_LEN];
   int   j, k;

   next_line (fp);

   /* read MCMC sample size */
   fscanf (fp, "%s %d", dummy, &data_pars_pt->smpl_size);
   next_line (fp);

   /* read gl_auto output order, by component (0) or not (1) */
   fscanf (fp, "%s %d", dummy, &data_pars_pt->by_component);
   next_line (fp);

   /* FIXME:
    * Shall we determine the MCMC sample size from the par file?
    * If so, should allow user to read only every #-th line using a option:
    *  "every  #".
    */

   /* data_pars_pt->smpl_size = NIt / FilIt; */

   /* Total number of meioses */

   data_pars_pt->n_components = nCompnt;
   data_pars_pt->n_meis_ped = (int *) malloc (sizeof (int) * nCompnt);
   data_pars_pt->n_meis = 0;
   for (k = 1; k <= nCompnt; k++)
   {
      data_pars_pt->n_meis_ped[k - 1] = cmplist[k].nmeioses;
      data_pars_pt->n_meis += cmplist[k].nmeioses;
   }

   /* Number of markers */
   /* NMarks: is the number of selected markers */
   /* Make sure to be consistent with gl_auto */
   data_pars_pt->n_loci = NMarks;

   /* Marker map */
   data_pars_pt->locus_map = (double *) malloc (sizeof (double)
                                                * data_pars_pt->n_loci);

   for (j = 0; j < data_pars_pt->n_loci; j++)
   {
      data_pars_pt->locus_map[j] = MarkPos[0][j];
   }

   /* TODO: how to handle male/female maps */

   /*
      fscanf(fp, "%s", dummy);
      for (j = 0; j < data_pars_pt->n_loci; j++) {
      fscanf(fp, "%lf", &data_pars_pt->locus_map[j]);
      }
      fscanf(fp, " \n");
    */
}

/**
 * Initialize trait model parameters (read some parameters from file).
 *
 * @param fp points to the input file
 *
 * @param trt_mod_pt points to the data structure storing the trait model parameters.
 *
 */
void read_trait_model_pars (FILE * fp, trait_model_t * trt_mod_pt)
{
   char  dummy[MAX_LEN], use_sample[MAX_LEN];
   double tmp;
   int   l_a_max = 16;                  /* FIXME better define it as a global? */

   /* Read stuff from the input file */
   next_line (fp);

   fscanf (fp, "%s %lf", dummy, &trt_mod_pt->pD);
   next_line (fp);

   /* FIXME: read pD from MORGAN? */
   /* trt_mod_pt->pD = PTraitPars-> */
   fscanf (fp, "%s %s", dummy, use_sample);
   next_line (fp);

   fscanf (fp, "%s %lf", dummy, &tmp);
   next_line (fp);

   if (use_sample[0] != 'y')
   {
      trt_mod_pt->mu = tmp;
   }

   fscanf (fp, "%s %s", dummy, use_sample);
   next_line (fp);

   if (use_sample[0] == 'y')
   {
      /* estimate se by sample sd */
      trt_mod_pt->se = trt_mod_pt->sd;
   }
   else
   {
      /* compute se based on specified additive and resid variance */
      trt_mod_pt->se = sqrt (PTraitPars->addi_var + PTraitPars->resi_var);
   }

   /*  Initialize other trait model parameters */
   trt_mod_pt->ca = (char *) malloc ((long) 1 << l_a_max);
   compute_ca (trt_mod_pt->ca, l_a_max);

   /* loddD is used when computing the w-score */
   trt_mod_pt->loddD = log (trt_mod_pt->pD / (1 - trt_mod_pt->pD));
}

/**
 * Read marginal test parameters
 *
 * @param fp points to the input file
 *
 * @miv_pars_pt points to the data structure for the marginal test parameters.
 */
void read_miv_pars (FILE * fp, miv_pars_t * miv_pars_pt,
                    data_pars_t * data_pars_pt)
{

   char  dummy[MAX_LEN];
   int   j;

   next_line (fp);                      /* Skip the first line. */

   fscanf (fp, "%s %s", dummy, miv_pars_pt->test);
   next_line (fp);

   fscanf (fp, "%s %d", dummy, &miv_pars_pt->n_mc);
   next_line (fp);

   fscanf (fp, "%s %d", dummy, &miv_pars_pt->n_pos);
   next_line (fp);

   fscanf (fp, "%s", dummy);
   miv_pars_pt->pos = (int *) malloc (sizeof (int) * miv_pars_pt->n_pos);
   for (j = 0; j < miv_pars_pt->n_pos; j++)
   {
      fscanf (fp, "%d", &miv_pars_pt->pos[j]);
   }
   next_line (fp);

   /* Assign map positions to the test positions */
   miv_pars_pt->map = (double *) malloc (sizeof (double) * miv_pars_pt->n_pos);

   for (j = 0; j < miv_pars_pt->n_pos; j++)
   {
      miv_pars_pt->map[j] = data_pars_pt->locus_map[miv_pars_pt->pos[j]];
   }
}

/**
 * Read conditional test parameters
 *
 * @param fp points to the input file
 *
 * @miv_pars_pt points to the data structure for the conditional test parameters.
 */
void read_civ_pars (FILE * fp, civ_pars_t * civ_pars_pt,
                    data_pars_t * data_pars_pt)
{

   char  dummy[MAX_LEN];
   int   j;

   next_line (fp);                      /* Skip a line. */

   fscanf (fp, "%s %s", dummy, civ_pars_pt->test);
   next_line (fp);

   fscanf (fp, "%s %d", dummy, &civ_pars_pt->n_mc);
   next_line (fp);

   fscanf (fp, "%s %d", dummy, &civ_pars_pt->n_pos);
   next_line (fp);

   civ_pars_pt->pos = (int *) malloc (sizeof (int) * civ_pars_pt->n_pos);
   civ_pars_pt->posl = (int *) malloc (sizeof (int) * civ_pars_pt->n_pos);
   civ_pars_pt->posr = (int *) malloc (sizeof (int) * civ_pars_pt->n_pos);
   fscanf (fp, "%s", dummy);
   for (j = 0; j < civ_pars_pt->n_pos; j++)
   {
      fscanf (fp, "%d", &civ_pars_pt->pos[j]);
   }
   next_line (fp);

   fscanf (fp, "%s", dummy);
   for (j = 0; j < civ_pars_pt->n_pos; j++)
   {
      fscanf (fp, "%d", &civ_pars_pt->posl[j]);
   }
   next_line (fp);

   fscanf (fp, "%s", dummy);
   for (j = 0; j < civ_pars_pt->n_pos; j++)
   {
      fscanf (fp, "%d", &civ_pars_pt->posr[j]);
   }
   next_line (fp);

   /* Assign map positions with the test and conditioning positions */
   civ_pars_pt->map = (double *) malloc (sizeof (double) * civ_pars_pt->n_pos);
   civ_pars_pt->mapl = (double *) malloc (sizeof (double) * civ_pars_pt->n_pos);
   civ_pars_pt->mapr = (double *) malloc (sizeof (double) * civ_pars_pt->n_pos);

   for (j = 0; j < civ_pars_pt->n_pos; j++)
   {
      civ_pars_pt->map[j] = data_pars_pt->locus_map[civ_pars_pt->pos[j]];
      civ_pars_pt->mapl[j] = data_pars_pt->locus_map[civ_pars_pt->posl[j]];
      civ_pars_pt->mapr[j] = data_pars_pt->locus_map[civ_pars_pt->posr[j]];
   }
}

/**
 * Record all possible test or conditioning positions.
 *
 * @param data_pars_pt points to the data set parameters
 * @param miv_pars_pt points to marginal test parameters
 * @param civ_pars_pt points to conditional test parameters
 */
void record_all_test_pos (data_pars_t * data_pars_pt,
                          const miv_pars_t * miv_pars_pt,
                          const civ_pars_t * civ_pars_pt)
{
   int   j, j1;                         /* index loci */
   int  *is_test_pos;

   is_test_pos = (int *) malloc (sizeof (int) * data_pars_pt->n_loci);
   for (j = 0; j < data_pars_pt->n_loci; j++)
   {
      is_test_pos[j] = 0;
   }

   for (j = 0; j < miv_pars_pt->n_pos; j++)
   {
      is_test_pos[miv_pars_pt->pos[j]] = 1;
   }

   for (j = 0; j < civ_pars_pt->n_pos; j++)
   {
      is_test_pos[civ_pars_pt->pos[j]] = 1;
      is_test_pos[civ_pars_pt->posl[j]] = 1;
      is_test_pos[civ_pars_pt->posr[j]] = 1;
   }

   data_pars_pt->n_pos = 0;
   for (j = 0; j < data_pars_pt->n_loci; j++)
   {
      data_pars_pt->n_pos += is_test_pos[j];
   }

   data_pars_pt->pos = (int *) malloc (sizeof (int) * data_pars_pt->n_pos);
   j1 = 0;
   for (j = 0; j < data_pars_pt->n_loci; j++)
   {
      if (is_test_pos[j])
      {
         data_pars_pt->pos[j1] = j;
         j1++;
      }
   }

   free (is_test_pos);
}

/**
 * Find the map positions corresponding to the test positions
 *
 * @param data_pars_pt points to the data set parameters
 * @param miv_pars_pt points to marginal test parameters
 * @param civ_pars_pt points to conditional test parameters
 */
/*
void map_test_pos(const data_pars_t *data_pars_pt, miv_pars_t *miv_pars_pt,
            civ_pars_t *civ_pars_pt) {
}
*/

/**
 * Read in the meiosis indicator matrix. In the input file, each row
 * corresponds to one meiosis in one simulation. Within each row, the first
 * entry is the meiosis indicator at an unlinked locus. The rest of the row
 * encodes meiosis indicators along a chromosome (marker map).  The second
 * entry specifies the meiosis indicator at the first locus of a marker map.
 * The third entry specifies how many switches(recombinations) are there in
 * this meiosis. The rest of the entries list all the positions at which
 * recombination (switch) happens. (This is equivalent to a run-length
 * encoding.)
 *
 * @param file_mis name of the input file.
 *
 * @param s_mat_pt points to a matrix that will be used to store the meiosis
 * indicators.
 *
 * @param s_unlinked_pt will be used to store meiosis indicators at the
 * unlinked locus.
 */

void read_mis (char *file_mis, mi_mat_t * s_mat_pt, mi_mat_t * s_unlinked_pt,
               data_pars_t * dat_pars_pt)
{
   int   smpl_size;
   int   n_meis;
   int   n_loci;
   /*  Will need a marker map */
   /*  Store the total number of pedigrees in the ped_list? */

   FILE *fp;
   int   h;                             /*indexes sample points */
   int   i;                             /*indexes meioses */
   int   j, j1, j2;                     /*index loci */
   int   b0, b1;                        /* meiosis indicators */
   int   n;                             /* number of switches */

   char  bit;

   typedef struct
   {
      char  first;                      /* meiosis indicator at the first locus */
      int   n_swis;                     /* number of switches */
      int  *swis;                       /* switch positions */
   } mic_t;                             /* a chain of meiosis indicators along a chromosome (same meiosis, different loci) */

   typedef struct
   {
      int   smpl_size;                  /* sample size */
      int   n_meis;                     /* number of meiosis indicator chains */
      int   n_loci;                     /* number of loci */
      mic_t **mics;                     /* meiosis indicator chains, each chain corresponds to a meiosis */
   } mic_list_t;                        /* list of meiosis indicator chains */

   mic_list_t mic_list;                 /* meiosis indicator chains at all meioses */

   smpl_size = dat_pars_pt->smpl_size;
   n_meis = dat_pars_pt->n_meis;
   n_loci = dat_pars_pt->n_loci;

   /*  -- Read the meiosis indicator matrix into <mic_list>. */
   mic_list.smpl_size = smpl_size;      /*?? */
   mic_list.n_meis = n_meis;            /*?? */
   mic_list.n_loci = n_loci;            /*?? */

   mic_list.mics = (mic_t **) malloc (sizeof (mic_t *) * smpl_size);
   for (h = 0; h < smpl_size; h++)
   {
      mic_list.mics[h] = (mic_t *) malloc (sizeof (mic_t) * n_meis);
   }

   s_unlinked_pt->smpl_size = smpl_size;
   s_unlinked_pt->n_meis = n_meis;
   s_unlinked_pt->n_loci = 1;

   s_unlinked_pt->s = (char ***) malloc (sizeof (char **) * smpl_size);
   s_unlinked_pt->map = (double *) malloc (sizeof (double));
   s_unlinked_pt->map[0] = 0;

   for (h = 0; h < smpl_size; h++)
   {
      s_unlinked_pt->s[h] = (char **) malloc (sizeof (char *) * n_meis);
      for (i = 0; i < n_meis; i++)
      {
         s_unlinked_pt->s[h][i] = (char *) malloc (sizeof (char)
                                                   * s_unlinked_pt->n_loci);
      }
   }

   if ((fp = fopen (file_mis, "r")) == NULL)
   {
      printf ("Cannot open file (meiosis indicators) %s for reading. \n",
              file_mis);
      exit (1);
   }

   for (h = 0; h < smpl_size; h++)
   {
      for (i = 0; i < n_meis; i++)
      {
         fscanf (fp, "%d %d %d", &b0, &b1, &n);
         s_unlinked_pt->s[h][i][0] = (char) b0;
         mic_list.mics[h][i].first = (char) b1;
         /*  fscanf(fp, "%d %d %d", &s_unlinked_pt->s[i], &mic_list.mics[i].first, &n); */
         mic_list.mics[h][i].n_swis = n;
         if (n > 0)
         {
            mic_list.mics[h][i].swis = (int *) malloc (sizeof (int) * n);
         }
         else
         {
            mic_list.mics[h][i].swis = NULL;
         }
         for (j = 0; j < n; j++)
         {
            fscanf (fp, "%d", &b1);
            mic_list.mics[h][i].swis[j] = b1 - 1;
            /*  Minus the positions by 1 since array starts with 0 in C. */
         }
      }
   }
   fclose (fp);

   /*  Check whether the chains of meiosis indicators are correctly read in. */
   for (h = 0; h < smpl_size; h++)
   {
      for (i = 0; i < n_meis; i++)
      {
         printf ("%d %d", mic_list.mics[h][i].first,
                 mic_list.mics[h][i].n_swis);
         for (j = 0; j < mic_list.mics[h][i].n_swis; j++)
         {
            printf (" %d", mic_list.mics[h][i].swis[j]);
         }
         printf ("\n");
      }
      printf ("\n");
   }

   /*  -- Convert the list of chains to a matrix representation */
   s_mat_pt->smpl_size = smpl_size;
   s_mat_pt->n_meis = n_meis;
   s_mat_pt->n_loci = n_loci;

   /*  -- Locus map */
   s_mat_pt->map = (double *) malloc (sizeof (double) * n_loci);
   for (j = 0; j < n_loci; j++)
   {
      s_mat_pt->map[j] = dat_pars_pt->locus_map[j];
   }

   s_mat_pt->s = (char ***) malloc (sizeof (char **) * smpl_size);
   for (h = 0; h < smpl_size; h++)
   {
      s_mat_pt->s[h] = (char **) malloc (sizeof (char *) * n_meis);
      for (i = 0; i < n_meis; i++)
      {
         s_mat_pt->s[h][i] = (char *) malloc (sizeof (char) * n_loci);
      }
   }

   for (h = 0; h < smpl_size; h++)
   {
      for (i = 0; i < n_meis; i++)
      {
         j = 0;
         bit = mic_list.mics[h][i].first;

         for (j1 = 0; j1 < mic_list.mics[h][i].n_swis; j1++)
         {
            for (j2 = j; j2 < mic_list.mics[h][i].swis[j1]; j2++)
            {
               s_mat_pt->s[h][i][j2] = bit;
            }
            bit = 1 - bit;
            j = j2;
         }

         /*  printf(": %d %d\n", j, mic_list.n_loci); */
         for (j2 = j; j2 < mic_list.n_loci; j2++)
         {
            s_mat_pt->s[h][i][j2] = bit;
         }
      }
   }

   /*  Free mic_list */
   for (h = 0; h < smpl_size; h++)
   {
      for (i = 0; i < mic_list.n_meis; i++)
      {
         free (mic_list.mics[h][i].swis);
      }
      free (mic_list.mics[h]);
   }
   free (mic_list.mics);
}

/**
 * Build the allele descent graph of the pedigree.
 *
 * @param ped_pt points to a pedigree
 *
 * @param adg_pt points to a allele descent graph
 *
 */
void query_ped_info (ped_t * ped_pt, adg_t * adg_pt)
{
   int   i;                             /*indexes individuals */
   int   i0, i1;                        /*indexes alleles; */
   int   iobs;                          /*indexes alleles; */
   int  *map;                           /* map allele indicies to inheritance vector indices */

   int  *n_offs, n_fndrs;
   int **i_offs, *i_fndrs;
   int   fixed_bit;

   /*  In order to satisfy the c89 standard for C programs, the following     */
   /*  variable, fixed_bits, was changed from a "long long" variable to just  */
   /*  a "long" variable.  As a result, we have only 32 bits available for    */
   /*  storing the fixed_bit_masks.  For this reason civil is limited to 15   */
   /*  founders, rather than the 30 total founders documented at the start of */
   /*  civil_datastruct.c.                                                    */

   unsigned long fixed_bits;

   strcpy (adg_pt->name, ped_pt->name);
   adg_pt->n_a = ped_pt->size * 2;

   /*  -- Count and index founder and nounder alleles. Indexes parental alleles */
   /*  of non-founder alleles. */

   /*  Count founder alleles */
   adg_pt->n_fa = 0;
   for (i = 0; i < ped_pt->size; i++)
   {
      adg_pt->n_fa += (ped_pt->inds[i].pa_id < 0) + (ped_pt->inds[i].ma_id < 0);
   }
   adg_pt->n_nfa = adg_pt->n_a - adg_pt->n_fa;

   adg_pt->i_fa = (int *) malloc (sizeof (int) * adg_pt->n_fa);
   adg_pt->i_nfa = (int *) malloc (sizeof (int) * adg_pt->n_nfa);
   /*  adg_pt->n_meis = adg_pt->n_nfa; */
   adg_pt->i_paa = (int *) malloc (sizeof (int) * adg_pt->n_nfa);
   adg_pt->i_maa = (int *) malloc (sizeof (int) * adg_pt->n_nfa);
   adg_pt->i_mi = (int *) malloc (sizeof (int) * adg_pt->n_nfa);

   map = (int *) malloc (sizeof (int) * adg_pt->n_a);

   i0 = 0;                              /* used to count founder alleles */
   i1 = 0;                              /* used to count nonfounde alleles */

   for (i = 0; i < ped_pt->size; i++)
   {
      if (ped_pt->inds[i].ma_id < 0)
      {
         adg_pt->i_fa[i0] = i * 2;
         i0++;
      }
      else
      {
         adg_pt->i_nfa[i1] = i * 2;
         map[i * 2] = i1;
         adg_pt->i_maa[i1] = ped_pt->inds[i].ma_id * 2;
         adg_pt->i_paa[i1] = adg_pt->i_maa[i1] + 1;
         adg_pt->i_mi[i1] = ped_pt->inds[i].ma_mi;
         i1++;
      }

      if (ped_pt->inds[i].pa_id < 0)
      {
         adg_pt->i_fa[i0] = i * 2 + 1;
         i0++;
      }
      else
      {
         adg_pt->i_nfa[i1] = i * 2 + 1;
         map[i * 2 + 1] = i1;
         adg_pt->i_maa[i1] = ped_pt->inds[i].pa_id * 2;
         adg_pt->i_paa[i1] = adg_pt->i_maa[i1] + 1;
         adg_pt->i_mi[i1] = ped_pt->inds[i].pa_mi;
         i1++;
      }
   }

   /* fprintf(stdmsg, "Done.\n"); */

   /*     Use the following block, if in the inheritance vector, we want to list the paternal */
   /*     meiosis indicator before the maternal meiosis indicator. */
   /*  */
   /*     i0 = 0; */
   /*     i1 = 0; */
   /*  */
   /*     for (i=0; i<ped_pt->size; i++) { */
   /*  if (ped_pt->inds[i].pa_id<0) { */
   /*      adg_pt->i_fa[i0] = i*2; */
   /*      i0++; */
   /*  } else { */
   /*      adg_pt->i_nfa[i1] = i*2; */
   /*      adg_pt->i_paa[i1] = ped_pt->inds[i].pa_id * 2; */
   /*      adg_pt->i_maa[i1] = adg_pt->i_paa[i1] + 1; */
   /*      adg_pt->i_mi[i1] = ped_pt->inds[i].pa_mi; */
   /*      i1++; */
   /*  } */
   /*  */
   /*  if (ped_pt->inds[i].ma_id<0) { */
   /*      adg_pt->i_fa[i0] = i*2+1; */
   /*      i0++; */
   /*  } else { */
   /*      adg_pt->i_nfa[i1] = i*2+1; */
   /*      adg_pt->i_paa[i1] = ped_pt->inds[i].ma_id * 2; */
   /*      adg_pt->i_maa[i1] = adg_pt->i_paa[i1] + 1; */
   /*      adg_pt->i_mi[i1] = ped_pt->inds[i].ma_mi; */
   /*      i1++; */
   /*  } */
   /*     } */
   /*  */

   /*  -- Count and index observed individuals */
   adg_pt->n_obs = 0;
   for (i = 0; i < ped_pt->size; i++)
   {
      adg_pt->n_obs += ped_pt->inds[i].obsd;
   }

   adg_pt->i_obsa = (int *) malloc (sizeof (int) * adg_pt->n_obs * 2);
   adg_pt->y = (double *) malloc (sizeof (double) * adg_pt->n_obs);

   iobs = 0;
   for (i = 0; i < ped_pt->size; i++)
   {
      if (ped_pt->inds[i].obsd)
      {
         adg_pt->y[iobs] = ped_pt->inds[i].y;
         adg_pt->i_obsa[iobs * 2] = i * 2;
         adg_pt->i_obsa[iobs * 2 + 1] = i * 2 + 1;
         iobs++;
      };
   }

   /*
      fprintf(stdmsg, "Figure out equivalent inheritance vector classes.\n");
    */

   /*  In a pedigree, due to founder phase uncertainty, probabilities (and */
   /*  w-scores, and bayes scores, etc.) of two inheritance vectors will be */
   /*  the same if the two inheritance vectors differ only in founder phase. */
   /*  */
   /*  For each founder in the pedigree, all his/her children will have a */
   /*  meiosis indicator corresponding to this founder. We put these meiosis */
   /*  indicators corresponding to a same founder parent in one group. In each */
   /*  group, we can hold one meiosis indicator fixed and not affect the */
   /*  calculations of likelihood, w-score, bayes scores etc. In other words, */
   /*  in an inheritance vector, if we flip ALL meiosis indicators in a */
   /*  meiosis group defined as above, we will get an equivalent inheritance */
   /*  vector. */
   /*  */
   /*  To uniquely identify a representative inheritance vector in each */
   /*  equivalence class, we can fix the the meiosis indicator of the first */
   /*  child of each founder parent at 0. */
   /*  */
   /*  Now we list which meiosis indicators can be fixed and generate bit */
   /*  masks that are useful for mapping a general inheritance vector to a */
   /*  representative inheritance vector in the same equivalende class: */
   /*  */
   /*    fixed_bit_masks: specify which bits can be fixed */
   /*  */
   /*    non_fixed_bit_maks: specify which bits are free to vary */
   /*  */
   /*    mei_grp_masks: specify which bits are in the same meiosis group, when */
   /*    we flip all bits within a same meiosis group, the resulted */
   /*    inheritance vector is equivalent to the orginal one. */

   /* / fprintf(stdout, "sizeof(fixed_bits) %d", sizeof(fixed_bits)); */
   /*  */
   /*  -- Generate masks. */

   /*  Build offspring lists for all individuals */
   n_offs = (int *) malloc (sizeof (int) * ped_pt->size);
   i_offs = (int **) malloc (sizeof (int *) * ped_pt->size);
   for (i = 0; i < ped_pt->size; i++)
   {
      n_offs[i] = 0;
      i_offs[i] = (int *) malloc (sizeof (int) * ped_pt->size);
   }

   for (i = 0; i < ped_pt->size; i++)
   {
      if (ped_pt->inds[i].ma_id >= 0)
      {
         i_offs[ped_pt->inds[i].ma_id][n_offs[ped_pt->inds[i].ma_id]] = i;
         n_offs[ped_pt->inds[i].ma_id]++;
      }
      if (ped_pt->inds[i].pa_id >= 0)
      {
         i_offs[ped_pt->inds[i].pa_id][n_offs[ped_pt->inds[i].pa_id]] = i;
         n_offs[ped_pt->inds[i].pa_id]++;
      }
   }

   /*  Index the founders in the pedigree */
   n_fndrs = 0;
   i_fndrs = (int *) malloc (sizeof (int) * ped_pt->size);
   for (i = 0; i < ped_pt->size; i++)
   {
      if ((ped_pt->inds[i].ma_id < 0) & (ped_pt->inds[i].pa_id < 0)
          & (n_offs[i] > 0))
      {
         i_fndrs[n_fndrs] = i;
         n_fndrs++;
      }
   }

   /*  Make sure that the total number of founders is not greater than 15.    */
   /*  This is necessary so as to not violate the c89 standard for C programs */
   /*  which we require within Morgan.  Without the following error check we  */
   /*  might, by mistake, shift a fixed_bit_mask past the left end of the     */
   /*  fixed_bits variable.                                                   */

   if (n_fndrs > 15)
   {
      give_err (ERQ_flm);
      give_up (ERR_qit);
   }

   /*  One meiosis indicator of the first child of each founder can be fixed. */
   adg_pt->n_fixed_bits = n_fndrs;
   adg_pt->n_non_fixed_bits = adg_pt->n_nfa - n_fndrs;

   adg_pt->fixed_bit_masks = (iv_t *) malloc (sizeof (iv_t) * n_fndrs);
   adg_pt->mei_grp_masks = (iv_t *) malloc (sizeof (iv_t) * n_fndrs);
   memset (adg_pt->mei_grp_masks, 0, sizeof (iv_t) * n_fndrs);
   adg_pt->non_fixed_bit_masks = (iv_t *) malloc (sizeof (iv_t)
                                                  * adg_pt->n_non_fixed_bits);

   /*  Generate fixed_bit_masks and mei_grp_masks */
   fixed_bits = 0;
   for (i = 0; i < n_fndrs; i++)
   {
      i0 = i_offs[i_fndrs[i]][0];       /* for the first child of this founder */
      if (ped_pt->inds[i0].ma_id == i_fndrs[i])
      {
         fixed_bit = i0 * 2;            /*maternal meiosis indicator can be fixed */
         for (i1 = 0; i1 < n_offs[i_fndrs[i]]; i1++)
         {
            adg_pt->mei_grp_masks[i] += 1 << map[i_offs[i_fndrs[i]][i1] * 2];
         }
      }
      else
      {
         fixed_bit = i0 * 2 + 1;        /*paternal meiosis indicator can be fixed */
         for (i1 = 0; i1 < n_offs[i_fndrs[i]]; i1++)
         {
            adg_pt->mei_grp_masks[i] += 1 << map[i_offs[i_fndrs[i]][i1] * 2
                                                 + 1];
         }
      }
      adg_pt->fixed_bit_masks[i] = 1 << map[fixed_bit];
      fixed_bits += (long) 1 << fixed_bit;
   }

   /*     print_iv(fixed_bits, adg_pt->n_nfa); */
   /*     printf("\n"); */
   /*
      fprintf(stdout, "%s: ", adg_pt->name);
      for (i = 0; i < adg_pt->n_a; i++) {
      fprintf(stdout, "%ld", (fixed_bits >> i) & 1);
      }
      fprintf(stdout, "\n");
    */

   /*  Generate non_fixed_bit_masks */
   i1 = 0;
   for (i = 0; i < adg_pt->n_nfa; i++)
   {
      if (((~fixed_bits) >> adg_pt->i_nfa[i]) & 1)
      {
         adg_pt->non_fixed_bit_masks[i1] = 1 << i;
         i1++;
      }
   }

   /*     print_iv(adg_pt->fixed_bit_masks, adg_pt->i_nfa); */
   /* fprintf(stdmsg, "Done.\n"); */

   /*  Free memories. */
   free (map);
   free (n_offs);
   free (i_fndrs);
   for (i = 0; i < ped_pt->size; i++)
   {
      free (i_offs[i]);
   }
   free (i_offs);
}

/**
 * Print parameters.
 */
void print_pars (file_names_t fns, data_pars_t data_pars, trait_model_t trt_mod,
                 miv_pars_t miv_pars, civ_pars_t civ_pars)
{
   int   j;

   /* Name of the meiosis indicator file. */
   fprintf (stdout, "Name of the meiosis indicator file output by gl_auto:\n");
   fprintf (stdout, "file_mis: %s\n", fns.file_mis);
   fprintf (stdout, "\n");

   /* Output file names */
   fprintf (stdout, "Output file directory and file names:\n");
   fprintf (stdout, "output dir: %s\n", fns.dir_out);
   fprintf (stdout, "file_miv_t: %s\n", fns.file_miv_t);
   fprintf (stdout, "file_miv_p: %s\n", fns.file_miv_p);
   fprintf (stdout, "file_civ_t: %s\n", fns.file_civ_t);
   fprintf (stdout, "file_civ_p: %s\n", fns.file_civ_p);
   fprintf (stdout, "\n");

   /* Info on the oscor file (inheritance vectors) */
   fprintf (stdout, "Inheritance vectors info:\n");
   fprintf (stdout, "n_mcmc: %d\n", data_pars.smpl_size);
   fprintf (stdout, "n_meis: %d\n", data_pars.n_meis);
   fprintf (stdout, "n_loci: %d\n", data_pars.n_loci);
   fprintf (stdout, "by_component: %d\n", data_pars.by_component);
   fprintf (stdout, "Marker map:");
   for (j = 0; j < data_pars.n_loci; j++)
   {
      fprintf (stdout, "%f ", data_pars.locus_map[j]);
   }
   fprintf (stdout, "\n\n");

   fprintf (stdout, "All possible test and conditioning positions:");
   for (j = 0; j < data_pars.n_pos; j++)
   {
      fprintf (stdout, "%d ", data_pars.pos[j]);
   }
   fprintf (stdout, "\n\n");

   /* Trait model parameters */
   fprintf (stdout, "Trait model parameters:\n");
   fprintf (stdout, "pD = %f\n", trt_mod.pD);
   fprintf (stdout, "mu = %f\n", trt_mod.mu);
   fprintf (stdout, "se = %f\n", trt_mod.se);
   fprintf (stdout, "\n");

   /* Marginal test parameters */
   fprintf (stdout, "Marginal test parameters:\n");
   fprintf (stdout, "test statistic: %s\n", miv_pars.test);
   fprintf (stdout, "n_mc: %d\n", miv_pars.n_mc);
   fprintf (stdout, "n_pos: %d\n", miv_pars.n_pos);
   fprintf (stdout, "Marginal test positions:");
   for (j = 0; j < miv_pars.n_pos; j++)
   {
      fprintf (stdout, "%d ", miv_pars.pos[j]);
   }
   fprintf (stdout, "\n\n");

   /* Conditional test parameters */
   fprintf (stdout, "Conditional test parameter:\n");
   fprintf (stdout, "test statistic: %s\n", civ_pars.test);
   fprintf (stdout, "n_mc: %d\n", civ_pars.n_mc);
   fprintf (stdout, "n_pos: %d\n", civ_pars.n_pos);
   fprintf (stdout, "Conditional test positions:\n");
   fprintf (stdout, "test_pos:  ");
   for (j = 0; j < civ_pars.n_pos; j++)
   {
      fprintf (stdout, "%d ", civ_pars.pos[j]);
   }
   fprintf (stdout, "\n");
   fprintf (stdout, "test_posl: ");
   for (j = 0; j < civ_pars.n_pos; j++)
   {
      fprintf (stdout, "%d ", civ_pars.posl[j]);
   }
   fprintf (stdout, "\n");
   fprintf (stdout, "test_posr: ");
   for (j = 0; j < civ_pars.n_pos; j++)
   {
      fprintf (stdout, "%d ", civ_pars.posr[j]);
   }
   fprintf (stdout, "\n\n");
}
