/* inline double w_qtl_sd(int * fgl, double* y, int n_obs, double loddD,
 *
 */

/**
 * Compute single-pedigree w-score as a function of observed trait values and
 * the vector of fgl labels:
 *
 *    w(fgl, y) = \sum_a Pr(y|g(fgl, a); \hat{mu}, sd) Pr(a)   ... (1)
 *
 * @param fgl a length <n_obs> x 2 vector, founder genome labels. The labels
 *   should start from 0.
 *
 * @param y a length <n_obs> vector, trait values.
 *
 * @param  n_obs number of observed trait values in the pedigree, length of vector <y>.
 *
 * @param  pD the allele frequency of D and
 *
 * @param loddD  log(pD/(1-pD)). If in the founder allelic type vector a, the number
 *   of D alleles is ca, then
 *
 *     log(Pr(a)) = ca * loddD + l_a * log(1 - p_D)  ... (2)
 *
 * @param  ca ca[i] is the number of 1's in the binary representation of i.
 *
 * @param  l_a the largest founder genome label in the vector <fgl>
 *
 * @return computed w-score
 *
 */
double w_qtl_sd (int *fgl, double *y, int n_obs, double loddD,
                 double pD, double sd, char *ca, int l_a)
{
   /*
    *  WARNING:
    *
    *    We assume that distinct founder alleles are labeled using 0, 1, 2,
    *    ...  The largest label appeared in <fgl> should be <l_a>. The time
    *    complexity of the code is of order:
    *
    *        n_obs x 2^l_a
    *
    *    So it is best to label the founder allelic types consecutively using
    *    0, 1, 2, ...  The code will break if <l_a> is greater than 31.
    *
    */

   double sd2, c1, c2;
   double p, sse;
   double sy[3];

   int   i, ig, D, ca2;
   int   ng[3];

   unsigned long a, n_a;

   long  mask;

   mask = (1 << 17) - 1;

   /*
    *  We have:
    *
    *    w(fgl, y) = \sum_a Pr(y|g(fgl, a); \hat{mu}, sd) Pr(a)   ... (1)
    *
    *    log(Pr(a)) = ca[a] * loddD + l_a * log(1 - p_D)                  ...(2)
    *
    * and for a specific g,
    *
    *    Pr(y | g; \hat{mu}, sd) = (2 * pi * sd^2)^{-n/2} exp ( - SSE / (2 * sd^2))   ... (3)
    *
    * wrhere n = n_obs and
    *
    *   sse = sse0 + sse1 + sse2.
    *
    * sse0 is the SSE for all individuals with dd genotype, etc,
    *
    *   sse = sum (y_i - mu_{g_i})^2.
    *       = sum(y_i)^2 - 1/n (sum(y_i))^2
    *
    */

   sd2 = sd * sd;
   c1 = exp (-n_obs / 2.0 * log (2.0 * M_PI * sd2) + l_a * log (1.0 - pD));
   /*
    * c2 = - 2.0 * sd2;
    * Downweight small pedigrees
    *
    */
   c2 = -2.0 * sd2 * n_obs / (n_obs + 3);

   /*
    * a: the bits of <a> represents the founder allelic types.
    *
    * D: the number of D alleles of each individual, this tells us the genotype of each individual.
    *
    */

   p = 0.0;
   n_a = 1 << l_a;

   for (a = 0; a < n_a; a++)
   {
      ig = 0;
      sy[0] = sy[1] = sy[2] = sse = 0;
      ng[0] = ng[1] = ng[2] = 0;

      for (i = 0; i < n_obs; i++)
      {
         D = (a >> fgl[ig++]) & 1;
         D += (a >> fgl[ig++]) & 1;

         ng[D]++;
         sy[D] += y[i];
         sse += y[i] * y[i];

      }

      for (i = 0; i < 3; i++)
      {
         if (ng[i] > 0)
         {
            sse -= sy[i] * sy[i] / ((double) ng[i]);
         }
      }
      ca2 = ca[a & mask] + ca[a >> 16]; /*a potential bug, double check! */
      p += exp (sse / c2 + ca2 * loddD);
   }

   return (p * c1);
}

/**
 * Compute w-scores for all pedigrees and store them to files.  As noted
 * earlier, two inheritance vectors are equivalent if they differ only by
 * founder phase. Accordingly, we can partition the set of all possible
 * inheritance vectors on each pedigree into equivalence classes. We will
 * choose one element from each equivalence class and compute w(s, y) for all
 * these representative inheritance vectors. We will store the computed
 * w-scores into files named after the corresponding names of the pedigrees.
 *
 * @param adg_list_t a list of allele descent graphs. An allele descent graph
 * is one abstract way to represent a pedigree.
 *
 * @param trt_mod trait model parameters
 *
 * @param dir output file directory
 */
void compute_lgw_tables (adg_list_t * adg_pt, trait_model_t * trt_mod,
                         char *dir)
{
   FILE *file_out;
   file_name_t file_name;

   iv_t  s, red_s;
   int   k;                             /* indexes pedigrees. */
   double *lgw;
   long  n_rep_s;

   int   fgl[128];                      /* FIXME: Potential bug */
   int   red_fgl[128];
   int   max_fgl;

   /* k = 0; */
   for (k = 0; k < adg_pt->n_peds; k++)
   {
      /*
         fprintf(stdmsg, "Processing pedigree %d ...\n", k);
       */
      /* FIXME: printlevel */
      /*
         fprintf(stdout, "adg graph %d: n_fa=%d, n_obs=%d, n_non_fixed_bits=%d\n",
         k,
         adg_pt->peds[k].n_fa, adg_pt->peds[k].n_obs,
         adg_pt->peds[k].n_non_fixed_bits);
       */

      sprintf (file_name, "%s/w/lgw.%s.out", dir, adg_pt->peds[k].name);

      /* fprintf(stdmsg, "%s\n", file_name); */

      file_out = fopen (file_name, "r"); /* check whether the file already exists */
      if (file_out == NULL)
      {                                 /* the file does not exist. */
         n_rep_s = 1 << adg_pt->peds[k].n_non_fixed_bits;
         lgw = (double *) malloc (sizeof (double) * n_rep_s);
         for (red_s = 0; red_s < n_rep_s; red_s++)
         {

            s = full_iv (red_s, &adg_pt->peds[k]);

            /*
             * print_iv(red_s, n_non_fixed_bits);
             * print_iv(s1, n_nfa);
             */

            iv2fgl (s, &adg_pt->peds[k], fgl);
            max_fgl = reduce_relabel_fgl (fgl, &adg_pt->peds[k], red_fgl);

            /*
             * print_fgl(fgl, n_nfa);
             *  print_fgl(red_fgl, n_obs*2 );
             */

            lgw[red_s] = log (w_qtl_sd (red_fgl, adg_pt->peds[k].y,
                                        adg_pt->peds[k].n_obs, trt_mod->loddD,
                                        trt_mod->pD, trt_mod->se, trt_mod->ca,
                                        max_fgl));

            /* printf("w(s) = %e\n", w_qtl_sd(red_fgl, y, n_obs, loddD, pD, sd, ca, max_fgl));
             * printf("w(s) = %e\n", w[red_s]);
             */
         }

         file_out = fopen (file_name, "wb");
         /* file_out = fopen("scores.out", "w");
          */
         if (file_out == NULL)
         {
            printf ("Can not open file %s for writing.\n", file_name);
            exit (1);
         }
         fwrite (lgw, sizeof (double), n_rep_s, file_out);
         fclose (file_out);
         free (lgw);

         fprintf (stdmsg, "created %s\n", file_name);
      }
      else
      {
         fprintf (stdmsg, "will use existing table %s \n", file_name);
         fclose (file_out);
      }
   }
}

/**
 * Read pre-computed w-scores from file. In future, check whether that file
 * exists, if it doesn't recompute?
 *
 * @param adg_pt points to a list of pedigrees
 *
 * @lgw_tabs_pt points to a structure that will be used to store the w-scores.
 *
 */
/* no longer needed
 *
void read_lgw_tables(const adg_list_t *adg_pt, lgw_tables_t *lgw_tabs_pt, char *dir) {
    FILE *file_out;
    file_name_t file_name;

    int k;

    iv_t s;

    lgw_tabs_pt->n_peds = adg_pt->n_peds;
    lgw_tabs_pt->n_ws = (long *) malloc(sizeof(long) * adg_pt->n_peds);
    lgw_tabs_pt->lgw = (double **) malloc(sizeof(double *) * adg_pt->n_peds);

    for (k=0; k<adg_pt->n_peds; k++) {
        lgw_tabs_pt->n_ws[k] =  1 << adg_pt->peds[k].n_non_fixed_bits;
        lgw_tabs_pt->lgw[k] = (double *) malloc(sizeof(double) * lgw_tabs_pt->n_ws[k]);

        sprintf(file_name, "%s/w/lgw.%s.out", dir, adg_pt->peds[k].name);
        file_out = fopen(file_name, "rb");
        if (file_out == NULL) {
            printf("Can not open file %s for reading.\n", file_name);
            exit(1);
        }
        fread(lgw_tabs_pt->lgw[k], sizeof(double), lgw_tabs_pt->n_ws[k], file_out);
        fclose (file_out);
    }
}

*/

/**
 * Read pre-computed w-scores from file. In future, check whether that file
 * exists, if it doesn't recompute?
 *
 * @param adg_pt points to a list of pedigrees
 *
 * @lgw_tabs_pt points to a structure that will be used to store the w-scores.
 *
 */
void read_lgw_table (const adg_list_t * adg_pt, lgw_table_t * lgw_tab_pt,
                     char *dir, int k)
{
   FILE *file_out;
   file_name_t file_name;

   lgw_tab_pt->n_ws = 1 << adg_pt->peds[k].n_non_fixed_bits;
   lgw_tab_pt->lgw = (double *) malloc (sizeof (double) * lgw_tab_pt->n_ws);

   sprintf (file_name, "%s/w/lgw.%s.out", dir, adg_pt->peds[k].name);
   file_out = fopen (file_name, "rb");
   if (file_out == NULL)
   {
      printf ("Can not open file %s for reading.\n", file_name);
      exit (1);
   }
   fread (lgw_tab_pt->lgw, sizeof (double), lgw_tab_pt->n_ws, file_out);
   fclose (file_out);
}

/**
 * Print w-score tables.  One table for each pedigree. Each table stores the
 * w-scores computed for all representative inheritance vectors for one
 * pedigree.
 *
 * @param lgw_tabs_pt points to the w-score tables.
 */
void print_lgw_tables (lgw_tables_t * lgw_tabs_pt)
{
   int   k;
   long  s;

   for (k = 0; k < lgw_tabs_pt->n_peds; k++)
   {
      printf ("\n ped %d\n", k);
      for (s = 0; s < lgw_tabs_pt->n_ws[k]; s++)
      {
         printf ("%e,   ", lgw_tabs_pt->lgw[k][s]);
      }
   }
}

/**
 * Marginal inheritance vector tests using w-score as the test statistic.
 *
 * @param adg_pt points to a list to pedigrees.
 *
 * @param s_pt points to sampled inheritance vectors.
 *
 * @param trt_mod trait model parameters. Need the following information:
 *   --pD: rare allele frequency
 *   --loddD:
 *   --ca:
 *   ...
 *
 * @param pars
 *
 */
void miv_w (const adg_list_t * adg_pt, const iv_list_t * s_pt,
            const trait_model_t * trt_mod_pt, miv_pars_t * pars_pt, char *dir)
{

   int   k;                             /*indexes pedigree */
   int   j;                             /*indexes loci */
   int   h, h1;                         /*indexes samples */
   int   n_sig;

   /* v_t rep_s, red_s; */
   iv_t  s1, *red_s_pt;

   iv_list_t s0_pt;

   double *t0;                          /*null w-scores.  */
   double t;

   lgw_table_t lgw_tab;

   /* Simulate null s'es */
   fprintf (stdmsg, "Simulate null inheritance vectors.\n");
   s0_pt.smpl_size = pars_pt->n_mc;
   s0_pt.n_loci = 1;
   s0_pt.map = NULL;

   simulate_rep_ivs (adg_pt, &s0_pt);

   /* Initialize variables */
   t0 = (double *) malloc (sizeof (double) * pars_pt->n_mc);
   for (h = 0; h < pars_pt->n_mc; h++)
   {
      t0[h] = 0;
   }

   pars_pt->smpl_size = s_pt->smpl_size;
   pars_pt->t = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);
   pars_pt->p = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);

   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      pars_pt->t[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      pars_pt->p[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         pars_pt->t[h][j] = 0;
      }
   }

   /* Compute w-socres by looking them up in a table */
   printf ("Compute w-scores.\n");

   for (k = 0; k < adg_pt->n_peds; k++)
   {
      /* read precomputed lgw_table for pedigree k */
      read_lgw_table (adg_pt, &lgw_tab, dir, k);
      /* Compute inheritance vector mapping table. */
      compute_iv_map (&adg_pt->peds[k], &red_s_pt);

      for (h = 0; h < pars_pt->n_mc; h++)
      {
         t0[h] += lgw_tab.lgw[s0_pt.s[h][0][k]];
      }

      for (h = 0; h < s_pt->smpl_size; h++)
      {
         for (j = 0; j < pars_pt->n_pos; j++)
         {
            /*               rep_s = rep_iv(s_pt->s[h][pars_pt->pos[j]][k], &adg_pt->peds[k]); */
            /*               red_s = reduced_iv(rep_s, &adg_pt->peds[k]); */
            /* Find the representative s corresponding to the observed s of
             * pedigree k, at test position j, and in MCMC sample h.
             */
            s1 = red_s_pt[s_pt->s[h][pars_pt->pos[j]][k]];
            pars_pt->t[h][j] += lgw_tab.lgw[s1];
         }
         /*     printf("\n"); */
      }

      free_lgw_table (&lgw_tab);
      free (red_s_pt);
   }

   free_iv_list (&s0_pt);

   /*
      printf("Compute p-values.\n");
      for (h=0; h<s_pt->smpl_size; h++) {
      for (j=0; j<s_pt->n_pos; j++) {
      t = pars_pt->t[h][j];
      printf("%8.5e, ", t);
      pars_pt->p[h][j] = (n_sig+1)/(pars_pt->n_mc+1);
      }
      printf("\n");
      }

      for (h1=0; h1<pars_pt->n_mc; h1++) {
      printf("%8.5e, ", t0[h1]);
      };
      printf("\n");
    */
   printf ("Compute p-values.\n");
   for (h = 0; h < s_pt->smpl_size; h++)
   {
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         n_sig = 0;
         t = pars_pt->t[h][j];
         for (h1 = 0; h1 < pars_pt->n_mc; h1++)
         {
            n_sig += (t0[h1] > t);
         };
         pars_pt->p[h][j] = ((double) n_sig + 1) / ((double) pars_pt->n_mc + 1);
      }
   }

   free (t0);
}

/**
 * Conditional inheritance vector tests using w-score as the test statistic.
 *
 * @param adg_pt points to a list to pedigrees.
 *
 * @param s_pt points to sampled inheritance vectors.
 *
 * @param trt_mod trait model parameters. Need the following information:
 *   --pD: rare allele frequency
 *   --loddD:
 *   --ca:
 *   ...
 *
 * @param pars
 *
 */
void civ_w (const adg_list_t * adg_pt, const iv_list_t * s_pt,
            const trait_model_t * trt_mod_pt, civ_pars_t * pars_pt, char *dir)
{

   int   i;                             /*indexes meioses */
   int   k;                             /*indexes pedigree */
   int   j;                             /*indexes loci */
   int   h;                             /*indexes the MCMC sample */
   int   h1;                            /*indexes the MC sample (for simulating null) */
   int   n_sig;
   int   n_meis;

   iv_t  sL, sR;
   double rL, rR;
   iv_t  s0;
   iv_t  rep_s, red_s, *red_s_pt;

   double p0, p1, p_tab[2][2];          /*  */
   double *p_sw;                        /* switching probabilities */

   int   bL, bR;

   double ***lgw0;
   double lgw;

   lgw_table_t lgw_tab;

   pars_pt->smpl_size = s_pt->smpl_size;

   /* Initialize variables */
   lgw0 = (double ***) malloc (sizeof (double **) * pars_pt->smpl_size);
   /* lgw0 takes a lot of memory! */
   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      lgw0[h] = (double **) malloc (sizeof (double *) * pars_pt->n_pos);
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         lgw0[h][j] = (double *) malloc (sizeof (double) * pars_pt->n_mc);
         for (h1 = 0; h1 < pars_pt->n_mc; h1++)
         {
            lgw0[h][j][h1] = 0;
         }
      }
   }

   pars_pt->t = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);
   pars_pt->p = (double **) malloc (sizeof (double *) * pars_pt->smpl_size);

   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      pars_pt->t[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      pars_pt->p[h] = (double *) malloc (sizeof (double) * pars_pt->n_pos);
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         pars_pt->t[h][j] = 0;
      }
   }

   /* Compute w-scores by looking them up in a table */
   fprintf (stdmsg, "Compute w-scores.\n");

   /* printf("Number of test positions: %d\n", pars_pt->n_pos); */

   for (k = 0; k < adg_pt->n_peds; k++)
   {
      fprintf (stdmsg, "Processing pedigree %d ...\n", k);
      fflush (stdmsg);
      /* read pre-computed table of log(w-score) for pedigree k */
      read_lgw_table (adg_pt, &lgw_tab, dir, k);

      /* compute a table that maps full iv to representative iv for pedigree k */
      compute_iv_map (&adg_pt->peds[k], &red_s_pt);

      n_meis = adg_pt->peds[k].n_nfa;
      p_sw = (double *) malloc (sizeof (double) * n_meis);

      /* Compute w-scores for observed s. */
      for (h = 0; h < s_pt->smpl_size; h++)
      {
         for (j = 0; j < pars_pt->n_pos; j++)
         {
            /* printf("(j, j1, j2) = (%d, %d, %d)\n",
             *    pars_pt->pos[j], pars_pt->posl[j], pars_pt->posr[j]);
             */
            /*
             * FIXME: could use red_s_pt to make this a litter faster
             */
            rep_s = rep_iv (s_pt->s[h][pars_pt->pos[j]][k], &adg_pt->peds[k]);
            red_s = reduced_iv (rep_s, &adg_pt->peds[k]);
            pars_pt->t[h][j] += lgw_tab.lgw[red_s];
         }
         /*     printf("\n"); */
      }

      /*    printf("(h, j, h1) = (%d, %d, %d)\n",  s_pt->smpl_size, pars_pt->n_pos, pars_pt->n_mc); */
      /* Compute w-score for conditional null s. */
      for (h = 0; h < s_pt->smpl_size; h++)
      {
         for (j = 0; j < pars_pt->n_pos; j++)
         {
            /* compute the probability of having a meiosis indicator 1 for all meiosis indicators at the test position. */
            /* FIXME: put into a seperate function? */
            sL = s_pt->s[h][pars_pt->posl[j]][k];
            sR = s_pt->s[h][pars_pt->posr[j]][k];
            rL = dis2rho (s_pt->map[pars_pt->pos[j]]
                          - s_pt->map[pars_pt->posl[j]]);
            rR = dis2rho (s_pt->map[pars_pt->posr[j]]
                          - s_pt->map[pars_pt->pos[j]]);

            for (bL = 0; bL < 2; bL++)
            {
               for (bR = 0; bR < 2; bR++)
               {
                  p1 = exp ((1 - bL) * log (rL) + bL * log (1 - rL) + (1
                                                                       -
                                                                       bR) *
                            log (rR) + bR * log (1 - rR));
                  p0 =
                     exp (bL * log (rL) + (1 - bL) * log (1 - rL) +
                          bR * log (rR) + (1 - bR) * log (1 - rR));
                  p_tab[bL][bR] = (p1 / (p0 + p1));
               }
            }

            for (i = 0; i < n_meis; i++)
            {
               p_sw[i] = p_tab[(sL >> i) & 1][(sR >> i) & 1];
            }

            for (h1 = 0; h1 < pars_pt->n_mc; h1++)
            {
               /* Simulate conditional null S */

               /*    printf("(h, j, h1) = (%d, %d, %d)\n", h, j, h1); */
               /*    printf("(rL, rR) = (%lf, %lf)\n",  */
               /*    (s_pt->map[pars_pt->pos[j]] - s_pt->map[pars_pt->posl[j]]), */
               /*    (s_pt->map[pars_pt->posr[j]] - s_pt->map[pars_pt->pos[j]])); */
               /*    printf("(rL, rR) = (%lf, %lf)\n", rL, rR); */

               /* s0 = simulate_linked_s(adg_pt->peds[k].n_nfa, sL, sR, p_tab); */
               s0 = simulate_binary_vec (n_meis, p_sw);

               /*                 printf("sL, sR, s0\n");
                  print_bits(sL, adg_pt->peds[k].n_nfa);
                  printf("\n");
                  print_bits(sR, adg_pt->peds[k].n_nfa);
                  printf("\n");
                  print_bits(s0, adg_pt->peds[k].n_nfa);
                  printf("\n");
                */
               /* rep_s = rep_iv(s0, &adg_pt->peds[k]); */
               /* red_s = reduced_iv(rep_s, &adg_pt->peds[k]); */
               /* lgw0[h][j][h1] += lgw_tab.lgw[red_s]; */
               lgw0[h][j][h1] += lgw_tab.lgw[red_s_pt[s0]];
            }
         }
      }

      /* free_lgw_tabs */
      free (red_s_pt);
      free_lgw_table (&lgw_tab);
      free (p_sw);
   }

   /*    free_lgw_tables(&lgw_tabs); */

   /*
      printf("Compute p-values.\n");
      for (h=0; h<pars_pt->smpl_size; h++) {
      for (j=0; j<pars_pt->n_pos; j++) {
      lgw = pars_pt->lgw[h][j];
      printf("%8.5e, ", lgw);
      pars_pt->p[h][j] = (n_sig+1)/(pars_pt->n_mc+1);
      for (h1=0; h1<pars_pt->n_mc; h1++) {
      printf("%8.5e, ", lgw0[h][j][h1]);
      };
      printf("\n");
      }
      printf("\n");
      }
    */

   printf ("Compute p-values.\n");
   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         n_sig = 0;
         lgw = pars_pt->t[h][j];
         for (h1 = 0; h1 < pars_pt->n_mc; h1++)
         {
            n_sig += (lgw0[h][j][h1] > lgw);
         };
         pars_pt->p[h][j] = ((double) n_sig + 1) / ((double) pars_pt->n_mc + 1);
      }
   }

   for (h = 0; h < pars_pt->smpl_size; h++)
   {
      for (j = 0; j < pars_pt->n_pos; j++)
      {
         free (lgw0[h][j]);
      }
      free (lgw0[h]);
   }
   free (lgw0);
}
