/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 AUTHOR
    Written by Yun Ju Sung.

 SEE ALSO
    README_twoqtl (in Lodscore directory) for user documentation.
    Lodscore/Gold directory to see examples.

 HISTORY
    Initial codes started from lm_markers.c (at 11/11/04).
    About 4000 lines of code (mostly in ../TwoQTL).  Apr  5, 2005
    Extended to do componentwise.                    May 20, 2005
    Cvs commit done.                                 Jul  5, 2005

    Use two streams of random numbers,
    one for segregation indicator and
    the other for founder gene alleles.              Aug  8, 2005

    Evaluate the null position more
    (to get more precision).                         Aug 18, 2005
*/

/**
 * @defgroup lm_twoqtl lm_twoqtl
 * @ingroup lodscore
 * @{
 */

#define PNAM  "lm_twoqtl"
#define TWOQTL

#include "lodscore.h"
#include "twoqtl.h"

double *set_theta ();
void  finishLodSc (void);

int main (int argc, char *argv[])
{
   mc_eval_t *MC_eval, *mc_eval;        /* MC evaluations for each batch.     */

   static FILE *oScorF = NULL;
   CTP_PTR p_ctp;
   RLD_PTR p_rld;
   MSD_PTR p_msd;

   char ***S_trait_p;                   /* Paternal SI at trait positions.    */
   char ***S_trait_m;                   /* Maternal SI at trait positions.    */

   double *mcmc_lod, *mcmc_sd;
   double **prob = NULL;
   double *theta = NULL;
   double *unused = NULL;

   int   nctemp;                        /* Backup copy of original n_compnts. */
   int   scored_iterations;             /* Number of scored iterations.       */
   int   bc;                            /* Current batch indicator.           */
   int   p1;

   int   nsw, start;
   int   cmp, ntp;
   int   iteration;
   int   mc_eval_offset;
   int   i1, i2;

   /* ................................  SET-UP   ............................ */

   nsw = setupAutozyg (argc, argv, PNAM, 2);
   setupSampling ();

   if (TraitSeedSet)
   {
      printf (" Hexadecimal trait seeds:");
      setsd_alt (TraitSeedC, TraitSeedT);
      printf ("  %#x %#x\n\n", TraitSeedC, TraitSeedT);
   }
   printf ("\n");

   /* Pass NTLocs as the argument to setupInherit.  This will be passed on as */
   /* the mm variable throughout the rest of the setup procedure.             */

   prob = setupInherit (NTLocs);

   if (MMSampler == Y)
   {
      /* Set up for the multiple-meiosis sampler.                             */
      Max_update = 8;
      P_meio = 0.2;
      P_ind = 0.3;
      P_sib = 0.3;
      P_3gen = 0.2;

      setup_marriage ();
      prnt_marriage ();
   }

   else if (MMSampler == N)
   {
      /* Set up for the old M-sampler.                                        */
      Max_update = 1;
      P_meio = 1.0;
      P_ind = 0.0;
      P_sib = 0.0;
      P_3gen = 0.0;
   }

   /* Copy the validated additive and residual variances to the SigAdd and    */
   /* SigErr global variables, respectively, so that the caller's values for  */
   /* these components of variance will be available throughout lm_twoqtl.    */

   SigAdd = P_trait_pars->addi_var;
   SigErr = P_trait_pars->resi_var;

   /* If gene-by-gene interaction is not present in the model, initialize     */
   /* the lm_twoqtl mean trait variables for each tloc for each genotype:     */
   /* geno_means_1[] and geno_means_2[].  Also initialize the vector of       */
   /* parameters, theta, for possible use in the future.                      */

   if (!GxGInteraction)
   {
      geno_means_1 = P_trait_pars->p_tloc_pars[0].means;
      if (NQtl == 2)
         geno_means_2 = P_trait_pars->p_tloc_pars[1].means;

      theta = set_theta ();
   }

   /* Now set evaluation points for lange_lik.                                */
   /* ntp is number of trait positions plus number of markers (NLocPos + 1).  */

   p_ctp = alloc_ctp (NLocPos + 1);

   ntp = set_eval_points (NMarks, p_ctp);

   if (ntp < NLocPos + 1)
      realloc_ctp (ntp, p_ctp);

   mcmc_lod = alloc_dbl1d (ntp);
   mcmc_sd = alloc_dbl1d (ntp);

   /* By setting cmp to 0, the lm_twoqtl program will consider all components */
   /* simultaneously while sampling.  Although the alternative of considering */
   /* each component separately while sampling is installed in this program,  */
   /* there is no capability to have lm_twoqtl invoke this alternative, since */
   /* cmp cannot be set to anything other than 0.                             */

   cmp = 0;

   /* Open output scores file, if requested to do so.                         */
   if (FilIt && DebugLevel == 1 && !oScorF)
      if (!(oScorF = open_fil (OSCORE)))
         give_up (ERR_qit);

   nctemp = n_compnts;
   if (cmp)
      n_compnts = 1;

   /* Call set_batch_size to make sure that the scoring interval evenly       */
   /* divides the realizations count and that the BatchSize global variable   */
   /* is set to a positive number that evenly divides the number of scored    */
   /* iterations.                                                             */

   set_batch_size ();
   printf ("\n");
   fflush (stdout);

   if (DebugLevel == 1)
   {
      /* Moved down initial peeling after alloc_rld etc: EAT July 2004.       */

      p_rld = alloc_rld ("CPBS", ntp, NBtch);

      set_simple_trans (loci, -1, -1, 0.5, 0.5, 0.5, 0.5);

      if (TraitTyp != QUANTITATIVE)
         p_rld->base_lik = locus_mjrpeel_cmp (cmp, 1.0, 1.0, 0, loci);
      else
         p_rld->base_lik = locus_mjrpeel_cmp (cmp, loci[0].ve, 1.0, 0, loci);

      /* Checks for zero probability in peeling operation.                    */
      if (p_rld->base_lik == -DBL_MAX)
         give_up_where (WHERE, ERS_zpp);

      printf (" Base log-likelihood log_10 P(Y_T) = %e\n",
              p_rld->base_lik / log (10.0));
      printf ("\n");
   }

   printf ("\n");
   fflush (stdout);

   /* Allocate the data structures needed for collecting the MCMC sampling    */
   /* statistical data.                                                       */

   p_msd = alloc_msd ();
   alloc_logprob ();

   MC_eval = alloc_mc_eval (NLocPos);

   ibase = get_ibase ();                /* Index of first ind in each comp.   */
   size_c = get_size_c ();              /* Number of ind in each component.   */
   obs_index = get_obs_index ();

   S_trait_p = alloc_S_trait (NLocPos, size_c);
   S_trait_m = alloc_S_trait (NLocPos, size_c);

   mbase = get_mbase (nsw);
   n_meiosis = get_n_meiosis (nsw);

   prntStatusBMC ();                    /* Print status for beginning MCMC.   */

   /* Setup variables for LM-multiple sampling: LPT Apr 2006.                 */
   setup_sampling_cmp (cmp, MMSampler, (NLoci - 1));

   /* ................  BURN-IN  .......................................      */

   /* Now do burn-in iterations.                                              */

   printf (" Now starting %d iterations of burn-in\n", NBrnIt);
   for (iteration = 1; iteration <= NBrnIt; ++iteration)
      sampling_cmp (iteration, cmp, NTLocs, MMSampler, (NLoci - 1), nFGL, prob);
   printf ("\n");
   fflush (stdout);

   set_evaluate ();                     /* Initialize various global          */
   /*                                      variables used by evaluate.        */

   /* ................  COMPUTE USING MCMC  ............................      */
   scored_iterations = 0;               /*Initialize scored iterations count. */
   start = 1;
   bc = 0;                              /* Set current batch indicator to 0.  */

   /* Now do the regular MCMC sampling, and score the results.                */

   printf (" Now starting %d iterations of MCMC\n", NIt);
   fflush (stdout);

   for (iteration = 1; (iteration <= NIt && bc < NBtch); ++iteration)
   {
      /* Now do the next MCMC sampling iteration.                             */

      sampling_cmp (iteration, cmp, NTLocs, MMSampler, (NLoci - 1), nFGL, prob);

      if (DebugMain)
      {
         fprintf (stderr, "\nAt iteration %d,\n", iteration);
         dbgXXGG ();
      }

      if ((iteration % ScorIt) == 0)
      {
         if (PrintLevel >= 4)
            update_counts ("", p_msd, cmp, 1, prob, unused);

         mc_eval_offset = n_compnts * bc;
         for (i1 = 0; i1 < n_compnts; ++i1)
         {
            update_S_trait (i1, nsw, S_trait_p, S_trait_m);

            /* Update batch sums at the alternative and null hypotheses.      */
            mc_eval = MC_eval + mc_eval_offset + i1;
            log_prob_alt (theta, 0, S_trait_p, S_trait_m, i1, nsw, start,
                          mc_eval);

            /* Update for the null hypothesis.                                */
            for (i2 = 0; i2 < MultiplierForNull; ++i2)
               log_prob_null (theta, 0, i1, nsw, start && (i2 == 0), mc_eval);

            if (DebugLevel == 4)
               print_mc_eval (MC_eval[mc_eval_offset + i1]);
         }

         ++scored_iterations;

         if (DebugLevel == 1)
         {
            /* Score rb_likelihoods.                                          */
            if (bc < NBtch)
               do_rblik_cmp (cmp, ntp, bc, p_ctp, p_rld);

            /* Every FilIt scored iterations, the current test positions log  */
            /* likelihoods need to be output to the interim scores file.      */

            /* Tests to see whether or not to output to scores file.          */
            if (FilIt && !(scored_iterations % FilIt))
            {
               fprintf (oScorF, "\n log-likelihoods: iter %d, cmpt %d\n",
                        iteration, cmp);
               for (p1 = 1; p1 < ntp; ++p1)
                  fprintf (oScorF, " %f", p_rld->cmpt_lik[p1]);
               fprintf (oScorF, "\n");
            }
         }

         /* Write batch means and start new batch if current batch is done.   */
         if (!(scored_iterations % BatchSize))
         {
            write_batch_means ((MC_eval + n_compnts * bc), bc);
            start = 1;
            ++bc;
         }
         else
            start = 0;

      }
   }

   /* Print status that we have finished with the MCMC.                       */
   printf ("\n\n");
   prntStatusFMC ();

   if (DebugLevel == 1)
   {
      mcmc_lods_results (p_rld, p_msd, p_ctp, ntp, cmp, scored_iterations,
                         mcmc_lod, mcmc_sd);

      printf ("\n\n === LodScore for all the components completed ===");
      printf ("\n\n");
      fflush (stdout);

      prntLodSHead ("LodScore estimates by Rao-Blackwellized computation");

      prntLodScores (ntp, mcmc_lod, mcmc_sd);

      free_rld (p_rld);
   }

   /* Compute lod scores and write them to the standard output file.          */
   write_lod_score (MC_eval);

   n_compnts = nctemp;

   if (FilIt && DebugLevel == 1 && (!cmp || (cmp == n_compnts)))
      fclose (oScorF);

   /* Free up allocated memory.                                               */
   free_sampling_cmp ();
   free_logprob ();
   free_msd (p_msd);

   free_1d ((void *) &mcmc_lod);
   free_1d ((void *) &mcmc_sd);
   free_ctp (p_ctp);

   free_update_S_trait ();
   if (SigAdd > 0.0)
      free_tq_gauss ();
   free_component ();
   free_evaluate ();

   free_S_trait (S_trait_p);
   free_S_trait (S_trait_m);

   free_mc_eval (MC_eval);
   free (theta);

   finishLodSc ();

   /* Now we are done.                                                        */

   exit (EXIT_SUCCESS);
}

double *set_theta ()
{
   double *theta;
   int   n_theta;

   n_theta = NQtl * 4 + 2;
   theta = (double *) malloc (n_theta * sizeof (double));
   alloc_chk (WHERE, theta);

   theta[0] = SigErr;
   theta[1] = SigAdd;
   theta[2] = Qq[0][1];
   memcpy (theta + 3, geno_means_1, 3 * sizeof (double));

   if (NQtl == 2)
   {
      theta[6] = Qq[1][1];
      memcpy (theta + 7, geno_means_2, 3 * sizeof (double));
   }

   return theta;
}

void finishLodSc (void)
{
   FILE *oSeedF;                        /* File pointer to the seed file.     */

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

   /* Free up allocated memory.                                               */
   free_xxgg ();
   free_allefrq ();
   free_chromo (0);
   free_allocate_loci ();
   free_perm ();
   locus_peel_free ();

   return;

}
