/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 NAME
    get_ibase, get_obs_ibase, get_size_c, get_obs_size_c, get_obs_index,
    get_mbase, get_n_meiosis, get_nFGL_c,
    free_component,
    nghd_index, comp_index,

 SYNOPSIS

    #include "twoqtl.h"
    int *get_ibase (void);
    int *get_obs_ibase (void);
    int *get_size_c (void);
    int *get_obs_size_c (void);
    int **get_obs_index (void);
    int *get_mbase (int nsw);
    int *get_n_meiosis (int nsw);
    int *get_nFGL_c (void);

    void free_component (void);

    int nghd_index (int i, int comp_no);
    int comp_index (int i, int comp_no);

 DESCRIPTION

    Get functions get variables for componentwise computation
    (allocating space if necessary).

    nghd_index gives the index in nghd of i-th ind in component comp_no
    and comp_index gives the index in component of i-th ind in nghd.

    free_component frees all the static data structures used here.

 SEE ALSO
    nghd_index and comp_index below

                                      documented  YJS  07/01/2005
 */

#include "twoqtl.h"

int  *get_ibase (void)
{
   int   comp_no = 0;
   int   i1;

   if (ibase == NULL)
   {
      ibase = (int *) malloc (n_compnts * sizeof (int));
      alloc_chk (WHERE, ibase);

      ibase[comp_no++] = 0;

      for (i1 = 0; i1 < pedsize; ++i1)
      {
         if (nghd[i1].component != comp_no)
            ibase[comp_no++] = i1;

         if (nghd[i1].component < comp_no)
            give_up_where (WHERE, ERR_qit);
      }
   }

   return ibase;
}

int  *get_obs_ibase (void)
{
   int   comp_no = 0, index = 0;
   int   i1;

   if (obs_ibase == NULL)
   {
      obs_ibase = (int *) malloc (n_compnts * sizeof (int));
      alloc_chk (WHERE, obs_ibase);

      obs_ibase[comp_no++] = index;

      for (i1 = 0; i1 < pedsize; ++i1)
         if (nghd[i1].q_dat != NULL)
         {
            if (nghd[i1].component != comp_no)
               obs_ibase[comp_no++] = index;

            if (nghd[i1].component < comp_no)
               give_up_where (WHERE, ERR_qit);

            ++index;
         }
   }

   return obs_ibase;
}

int **get_obs_index (void)
{
   int  *o_index, comp_size, j_nghd;
   int   index, i1, i2;

   if (obs_index == NULL)
   {
      obs_index = (int **) malloc (n_compnts * sizeof (int *));
      alloc_chk (WHERE, obs_index);

      if (obs_size_c == NULL)
         obs_size_c = get_obs_size_c ();
      if (size_c == NULL)
         size_c = get_size_c ();

      for (i1 = 0; i1 < n_compnts; ++i1)
      {
         obs_index[i1] = (int *) malloc (obs_size_c[i1] * sizeof (int));
         alloc_chk (WHERE, obs_index[i1]);

         o_index = obs_index[i1];
         comp_size = size_c[i1];

         for (i2 = index = 0; i2 < comp_size; ++i2)
         {
            j_nghd = nghd_index (i2, i1);

            if (nghd[j_nghd].q_dat != NULL)
               o_index[index++] = i2;
         }
      }
   }

   return obs_index;
}

int  *get_size_c (void)
{

   if (size_c == NULL)
   {
      if (ibase == NULL)
         ibase = get_ibase ();
      size_c = construct_size_c (ibase, pedsize);
   }

   return size_c;
}

int  *get_obs_size_c (void)
{

   if (obs_size_c == NULL)
   {
      if (obs_ibase == NULL)
         obs_ibase = get_obs_ibase ();
      obs_size_c = construct_size_c (obs_ibase, Nq);
   }

   return obs_size_c;
}

int  *construct_size_c (int *base, int max_base)
{
   int  *new_size;
   int   i1;

   new_size = (int *) malloc (n_compnts * sizeof (int));
   alloc_chk (WHERE, new_size);

   for (i1 = 0; i1 < n_compnts; i1++)
   {
      new_size[i1] = (i1 == n_compnts - 1) ? max_base : base[i1 + 1];
      new_size[i1] -= base[i1];
   }

   return new_size;
}

int  *get_mbase (int nsw)
{
   int   comp_no = 0;
   int   i1;

   if (mbase == NULL)
   {
      mbase = (int *) malloc (n_compnts * sizeof (int));
      alloc_chk (WHERE, mbase);

      mbase[comp_no++] = 1;             /* Meiosis is 1-origin indexing.      */

      for (i1 = 1; i1 <= nsw; ++i1)
      {
         if (meiosis[i1].cmpnt != comp_no)
            mbase[comp_no++] = i1;

         if (meiosis[i1].cmpnt < comp_no)
            give_up_where (WHERE, ERR_qit);
      }

      for (i1 = 0; i1 < n_compnts; ++i1) /* Compare with cmplist.             */
         if (mbase[i1] != cmplist[i1 + 1].meioses[0])
            give_up_where (WHERE, ERR_qit);
   }

   return mbase;
}

int  *get_n_meiosis (int nsw)
{
   int   i1;

   if (n_meiosis == NULL)
   {
      if (mbase == NULL)
         mbase = get_mbase (nsw);

      n_meiosis = construct_size_c (mbase, nsw + 1);

      for (i1 = 0; i1 < n_compnts; ++i1) /* Compare with cmplist.             */
         if (n_meiosis[i1] != cmplist[i1 + 1].nmeioses)
            give_up_where (WHERE, ERR_qit);
   }

   return n_meiosis;
}

int  *get_nFGL_c (void)
{
   int   comp_no = 0, count = 0;
   int   i1;

   if (nFGL_c == NULL)
   {
      nFGL_c = (int *) malloc (n_compnts * sizeof (int));
      alloc_chk (WHERE, nFGL_c);

      for (i1 = 0; i1 < pedsize; ++i1)
      {
         if (nghd[i1].component != comp_no + 1)
         {
            nFGL_c[comp_no++] = count * 2;
            count = 0;
         }
         if (nghd[i1].pa->index == INVALID_INDEX)
            ++count;
      }
      nFGL_c[comp_no] = count * 2;

      for (i1 = 0; i1 < n_compnts; ++i1)
         if (nFGL_c[i1] != (cmplist[i1 + 1].nfou) * 2)
            give_up_where (WHERE, ERR_qit);
   }

   return nFGL_c;
}

/* Map i-th individual in component comp_no to index in nghd and check that   */
/* component of this individual is comp_no.  Note: zero-origin indexing at    */
/* both ends.                                                                 */

int nghd_index (int i, int comp_no)
{
   int   k;

   if (ibase == NULL)
      ibase = get_ibase ();

   k = (i == INVALID_INDEX) ? i : i + ibase[comp_no];

   if (nghd[k].component != comp_no + 1)
      give_up_where (WHERE, ERR_qit);

   return k;
}

/* Map i-th individual in nghd to index in component comp_no.
 *
 * Without error:
 *   comp_index (nghd_index (i, comp_no), comp_no) == i
 *   nghd_index (comp_index (j, comp_no), comp_no) == j                       */

int comp_index (int i, int comp_no)
{
   int   k;

   if (ibase == NULL)
      ibase = get_ibase ();

   k = (i == INVALID_INDEX) ? i : i - ibase[comp_no];

   return k;
}

void free_component (void)
{
   int   i1;

   if (ibase != NULL)
      free (ibase);
   if (size_c != NULL)
      free (size_c);
   if (obs_ibase != NULL)
      free (obs_ibase);
   if (obs_size_c != NULL)
      free (obs_size_c);
   if (obs_index != NULL)
   {
      for (i1 = 0; i1 < n_compnts; ++i1)
         free (obs_index[i1]);
      free (obs_index);
   }
   if (mbase != NULL)
      free (mbase);
   if (n_meiosis != NULL)
      free (n_meiosis);
   if (nFGL_c != NULL)
      free (nFGL_c);

   return;
}
